import json
from typing import Any, Dict, List, Optional

import boto3
import numpy as np
from botocore.config import Config

DYNAMODB = "dynamodb"
TABLE = "Table"
ATTRIBUTE_NAME = "AttributeName"
ATTRIBUTE_TYPE = "AttributeType"
KEY_TYPE = "KeyType"
KEY_SCHEMA = "KeySchema"
KEYS = "Keys"
ITEM = "Item"
ITEMS = "Items"
DYNAMODB_STRING_TYPE = "S"
DYNAMODB_NUMBER_TYPE = "N"
HASH = "HASH"
RANGE = "RANGE"
BATCH_GET_ITEM_LIMIT = 100
ATTRIBUTES_TO_GET = "AttributesToGet"

PRIMARY_KEY_ATTRIBUTE_NAME_VALUE = "_feature_store_internal__primary_keys"
PRIMARY_KEY_SCHEMA = {ATTRIBUTE_NAME: PRIMARY_KEY_ATTRIBUTE_NAME_VALUE, KEY_TYPE: HASH}
TTL_KEY_ATTRIBUTE_NAME_VALUE = "_feature_store_internal__ttl_column"

DEFAULT_CONFIG = Config(
    retries={"max_attempts": 3, "mode": "standard"}, read_timeout=30
)


def to_dynamodb_primary_key(primary_key_values: List[Any]):
    return {PRIMARY_KEY_ATTRIBUTE_NAME_VALUE: json.dumps(primary_key_values)}


def to_range_schema(timestamp_key: str):
    return {ATTRIBUTE_NAME: timestamp_key, KEY_TYPE: RANGE}


def key_schema_to_tuple(key_schema_element: Dict[str, str]):
    return (key_schema_element[ATTRIBUTE_NAME], key_schema_element[KEY_TYPE])


def key_schemas_equal(key_schema1: List[Dict], key_schema2: List[Dict]):
    # Compare two KeySchemas, ignoring order of items.
    key_schema_tuples1 = set(key_schema_to_tuple(ks) for ks in key_schema1)
    key_schema_tuples2 = set(key_schema_to_tuple(ks) for ks in key_schema2)
    return key_schema_tuples1 == key_schema_tuples2


def to_safe_select_expression(feature_names: List[str]):
    """
    Helper to create the args for safe feature selection in DynamoDB. DynamoDB projection expressions (which define the
    attributes to retrieve), cannot be used with reserved keywords or special characters (including spaces).
    e.g. "feat 1", "_feat_1", "comment" are all invalid projection expressions.

    See the Amazon documentation for more details:
    https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_Query.html#DDB-Query-request-ExpressionAttributeNames

    To be safe, we alias all features with [#f0, #f1...]. Selecting ["feat 1", "_feat_1", "comment"] yields:
        - ProjectionExpression="#f0, #f1, #f2"
        - ExpressionAttributeNames={"#f0": "feat 1", "#f1": "_feat_1", "#f2": "comment}

    :param feature_names: List of feature names to select.
    :return: Safe DynamoDB expression args for selecting the given features.
    """
    safe_aliases = [f"#f{i}" for i in range(len(feature_names))]
    return {
        "ProjectionExpression": ", ".join(safe_aliases),
        "ExpressionAttributeNames": dict(zip(safe_aliases, feature_names)),
    }


def get_dynamodb_resource(
    access_key_id: str,
    secret_access_key: str,
    region: str,
    session_token: Optional[str] = None,
    endpoint_url: Optional[str] = None,
):
    return boto3.resource(
        DYNAMODB,
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        region_name=region,
        aws_session_token=session_token,
        endpoint_url=endpoint_url,
        config=DEFAULT_CONFIG,
    )


def paginate_keys(
    batch_keys: Dict[str, List[Any]],
    batch_size_limit: int,
) -> List[Dict[str, List[Any]]]:
    """
    :param batch_keys: A large dictionary from key to list of items.
    :param batch_size_limit:  item size limit for each dictionary
    :return a list of smaller dictionaries with item size not exceeed the limit
    """
    result = []
    count = 0
    for table in batch_keys.keys():
        # if count is 0 (using new mini batch) create a new dictionary
        if count == 0:
            result.append({table: []})
        # otherwise continue filling up most recent dictionary with table key until batch limit hit
        else:
            result[-1][table] = []

        num_items = len(batch_keys[table])
        # if current count + num_items is less than batch limit, append keys for table to current dictionary
        if count + num_items <= batch_size_limit:
            result[-1][table].extend(batch_keys[table])
        # otherwise append until limit reached and create new dictionaries for remaining items in table
        else:
            batch_remaining = batch_size_limit - count
            result[-1][table].extend(batch_keys[table][:batch_remaining])
            for i in range(batch_remaining, num_items, batch_size_limit):
                result.append(
                    {table: batch_keys[table][i : min(num_items, i + batch_size_limit)]}
                )

        # update count within dictionary, mod by batch_size_limit to reset on each new dictionary
        count = (count + num_items) % batch_size_limit

    return result


def merge_batched_results(
    batched_result: List[Dict[str, List[Any]]]
) -> Dict[str, List[Any]]:
    """
    :param batched_result: A list of small dictionaries.
    :return One single dictionary that combines the small dictionaries.
    """
    result = {}
    for batch in batched_result:
        for table in batch.keys():
            if table not in result:
                result[table] = []
            result[table].extend(batch[table])

    return result



""" Defines the LookupEngine class, which is used to perform lookups on online stores. This class
differs from Publish in that its actions are read-only.
"""

import abc
import collections
import functools
import logging
from typing import Any, Dict, List, Tuple

import numpy as np
import pandas as pd

from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.feature_store.utils.metrics_utils import LookupClientMetrics
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)
from databricks.ml_features_common.entities.online_store_for_serving import (
    MySqlConf,
    SqlServerConf,
)

LookupKeyType = Tuple[str, ...]


class LookupEngine(abc.ABC):
    @abc.abstractmethod
    def lookup_features(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        raise NotImplementedError

    @abc.abstractmethod
    def batch_lookup_features(
        self,
        lookup_list_dict: Dict[str, Dict[LookupKeyType, LookupKeyList]],
        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
        *,
        metrics: LookupClientMetrics = None,
    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
        raise NotImplementedError

    @abc.abstractmethod
    def shutdown(self) -> None:
        raise NotImplementedError



import warnings
from collections import Counter
from functools import wraps
from typing import Any, List

from mlflow.store.artifact.models_artifact_repo import ModelsArtifactRepository
from mlflow.store.artifact.runs_artifact_repo import RunsArtifactRepository


def is_artifact_uri(uri):
    """
    Checks the artifact URI is associated with a MLflow model or run.
    The actual URI can be a model URI, model URI + subdirectory, or model URI + path to artifact file.
    """
    return ModelsArtifactRepository.is_models_uri(
        uri
    ) or RunsArtifactRepository.is_runs_uri(uri)


def get_duplicates(elements: List[Any]) -> List[Any]:
    """
    Returns duplicate elements in the order they first appear.
    """
    element_counts = Counter(elements)
    duplicates = []
    for e in element_counts.keys():
        if element_counts[e] > 1:
            duplicates.append(e)
    return duplicates


def validate_strings_unique(strings: List[str], error_template: str):
    """
    Validates all strings are unique, otherwise raise ValueError with the error template and duplicates.
    Passes single-quoted, comma delimited duplicates to the error template.
    """
    duplicate_strings = get_duplicates(strings)
    if duplicate_strings:
        duplicates_formatted = ", ".join([f"'{s}'" for s in duplicate_strings])
        raise ValueError(error_template.format(duplicates_formatted))


def get_unique_list_order(elements: List[Any]) -> List[Any]:
    """
    Returns unique elements in the order they first appear.
    """
    return list(dict.fromkeys(elements))


_DEPRECATED_MARK_ATTR_NAME = "__deprecated"


def mark_deprecated(func):
    """
    Mark a function as deprecated by setting a private attribute on it.
    """
    setattr(func, _DEPRECATED_MARK_ATTR_NAME, True)


def is_marked_deprecated(func):
    """
    Is the function marked as deprecated.
    """
    return getattr(func, _DEPRECATED_MARK_ATTR_NAME, False)


def deprecated(alternative=None, since=None, impact=None):
    """
    Annotation decorator for marking APIs as deprecated in docstrings and raising a warning if
    called.
    :param alternative: (Optional string) The name of a superseded replacement function, method,
                        or class to use in place of the deprecated one.
    :param since: (Optional string) A version designator defining during which release the function,
                  method, or class was marked as deprecated.
    :param impact: (Optional boolean) Indication of whether the method, function, or class will be
                   removed in a future release.
    :return: Decorated function.

    TODO[ML-35177]: Migrate change back to mlflow. Changes:
      - Marks the decorated function with attribute `__deprecated`
    """

    def deprecated_decorator(func):
        since_str = " since %s" % since if since else ""
        impact_str = (
            impact if impact else "This method will be removed in a future release."
        )

        notice = (
            "``{qual_function_name}`` is deprecated{since_string}. {impact}".format(
                qual_function_name=".".join([func.__module__, func.__qualname__]),
                since_string=since_str,
                impact=impact_str,
            )
        )
        if alternative is not None and alternative.strip():
            notice += " Use ``%s`` instead." % alternative

        @wraps(func)
        def deprecated_func(*args, **kwargs):
            warnings.warn(notice, category=FutureWarning, stacklevel=2)
            return func(*args, **kwargs)

        if func.__doc__ is not None:
            deprecated_func.__doc__ = ".. Warning:: " + notice + "\n" + func.__doc__

        mark_deprecated(deprecated_func)

        return deprecated_func

    return deprecated_decorator



import logging


def get_logger(name: str, log_level: int = logging.WARNING) -> logging.Logger:
    logger = logging.getLogger(name)
    # A handler needs to be created to config the current logger.
    if not logger.handlers:
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(levelname)s %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    logger.setLevel(log_level)
    # Disable propagation to prevent duplicate logging when root logger has handlers
    logger.propagate = False
    return logger



from urllib3.util import Retry

from databricks.feature_store.utils.logging_utils import get_logger

_logger = get_logger(__name__)


class RetryWithLogging(Retry):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def new(self, **kwargs) -> "RetryWithLogging":
        return super().new(**kwargs)

    def increment(
        self,
        method=None,
        url=None,
        response=None,
        error=None,
        _pool=None,
        _stacktrace=None,
    ) -> "RetryWithLogging":
        _logger.warning(
            f"Retrying request: {method} {url} due to http status: {response.status if response else None}"
        )
        return super().increment(method, url, response, error, _pool, _stacktrace)



from typing import List, Tuple

import pandas as pd

from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
    PrimaryKeyDetails,
)
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    DataType as ProtoDataType,
)

LookupKeyType = Tuple[str, ...]


def validate_lookup_key_datatype(
    df: pd.DataFrame,
    primary_keys: List[PrimaryKeyDetails],
    lookup_key: LookupKeyType,
):
    """
    Validates that the given keys are of the correct datatype used to look up from online store.
    It is assumed that lookup_key and primary_keys are in the same order.
    """
    dtypes_dict = df.dtypes.to_dict()
    for pk, lookup_name in zip(primary_keys, lookup_key):
        if pk.data_type == ProtoDataType.STRING and not pd.api.types.is_string_dtype(
            dtypes_dict[lookup_name]
        ):
            raise ValueError(f"Provided primary keys {pk.name} is not of type string")


def get_primary_key_list(
    lookup_key: LookupKeyType,
    primary_keys: List[PrimaryKeyDetails],
    df: pd.DataFrame,
) -> LookupKeyList:
    """
    :return: A list of rows where each row is a list of tuples, each tuple is (column_name, value).
        There will be a column per column in `primary_keys`, and a row per row in `df`. `lookup_key` must `primary_keys` in length.
        The order of columns in each row is the same as the order of columns in `primary_keys`.
        Note that using LookupKeyList (which uses Python lists) is faster than using pandas
        DataFrame for the majority of use cases, since pandas DataFrame is optimized for large
        number of rows and has heavy overhead for small data.
        It is assumed that lookup_key and primary_keys are in the same order.
    """
    lookup_key_columns = df[list(lookup_key)]
    validate_lookup_key_datatype(lookup_key_columns, primary_keys, lookup_key)
    lookup_key_dict = lookup_key_columns.to_dict("list")
    # Update the lookup_key_df column names to be those of the feature table primary
    # key columns, rather than the names of columns from the source DataFrame. This is
    # required by the lookup_features interface.
    return LookupKeyList(
        columns=[feature_pk.name for feature_pk in primary_keys],
        rows=[
            [lookup_key_dict[lookup_name][i] for lookup_name in lookup_key]
            for i in range(len(df))
        ],
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: feature_spec.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    4,
    '',
    'feature_spec.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from databricks.ml_features_common.protos import feature_pb2 as feature__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x66\x65\x61ture_spec.proto\x12\x12\x66\x65\x61turestorecommon\x1a\x15scalapb/scalapb.proto\x1a\rfeature.proto\"$\n\x14SourceDataColumnInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x9b\x01\n\x11\x46\x65\x61tureColumnInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x14\n\x0c\x66\x65\x61ture_name\x18\x02 \x01(\t\x12\x12\n\nlookup_key\x18\x03 \x03(\t\x12\x13\n\x0boutput_name\x18\x04 \x01(\t\x12\x1c\n\x14timestamp_lookup_key\x18\x05 \x03(\t\x12\x15\n\rdefault_value\x18\x06 \x01(\t\"3\n\x0cInputBinding\x12\x11\n\tparameter\x18\x01 \x01(\t\x12\x10\n\x08\x62ound_to\x18\x02 \x01(\t\"u\n\x12OnDemandColumnInfo\x12\x10\n\x08udf_name\x18\x01 \x01(\t\x12\x38\n\x0einput_bindings\x18\x02 \x03(\x0b\x32 .featurestorecommon.InputBinding\x12\x13\n\x0boutput_name\x18\x03 \x01(\t\"\xb8\x02\n\nColumnInfo\x12K\n\x17source_data_column_info\x18\x01 \x01(\x0b\x32(.featurestorecommon.SourceDataColumnInfoH\x00\x12\x44\n\x13\x66\x65\x61ture_column_info\x18\x02 \x01(\x0b\x32%.featurestorecommon.FeatureColumnInfoH\x00\x12G\n\x15on_demand_column_info\x18\x03 \x01(\x0b\x32&.featurestorecommon.OnDemandColumnInfoH\x00\x12\x15\n\x07include\x18\n \x01(\x08:\x04true\x12\x11\n\tdata_type\x18\x0b \x01(\t\x12\x1c\n\x14topological_ordering\x18\x0c \x01(\x05\x42\x06\n\x04info\"Q\n\x10\x46\x65\x61tureTableInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x10\n\x08table_id\x18\x02 \x01(\t\x12\x17\n\x0flookback_window\x18\x03 \x01(\x01\" \n\x0c\x46unctionInfo\x12\x10\n\x08udf_name\x18\x01 \x01(\t\"\xc5\x02\n\x0b\x46\x65\x61tureSpec\x12\x35\n\rinput_columns\x18\x01 \x03(\x0b\x32\x1e.featurestorecommon.ColumnInfo\x12\x1d\n\x15serialization_version\x18\x02 \x01(\x05\x12\x14\n\x0cworkspace_id\x18\x03 \x01(\x03\x12$\n\x1c\x66\x65\x61ture_store_client_version\x18\x04 \x01(\t\x12:\n\x0cinput_tables\x18\x05 \x03(\x0b\x32$.featurestorecommon.FeatureTableInfo\x12\x39\n\x0finput_functions\x18\x06 \x03(\x0b\x32 .featurestorecommon.FunctionInfo\x12-\n\x08\x66\x65\x61tures\x18\x07 \x03(\x0b\x32\x1b.featurestorecommon.FeatureBu\n.com.databricks.proto.featurestorecommon.jprotoB\x10\x46\x65\x61tureSpecProtoP\x01\xa0\x01\x01\xe2?+\n\'com.databricks.proto.featurestorecommon\x10\x01')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feature_spec_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n.com.databricks.proto.featurestorecommon.jprotoB\020FeatureSpecProtoP\001\240\001\001\342?+\n\'com.databricks.proto.featurestorecommon\020\001'
  _globals['_SOURCEDATACOLUMNINFO']._serialized_start=80
  _globals['_SOURCEDATACOLUMNINFO']._serialized_end=116
  _globals['_FEATURECOLUMNINFO']._serialized_start=119
  _globals['_FEATURECOLUMNINFO']._serialized_end=274
  _globals['_INPUTBINDING']._serialized_start=276
  _globals['_INPUTBINDING']._serialized_end=327
  _globals['_ONDEMANDCOLUMNINFO']._serialized_start=329
  _globals['_ONDEMANDCOLUMNINFO']._serialized_end=446
  _globals['_COLUMNINFO']._serialized_start=449
  _globals['_COLUMNINFO']._serialized_end=761
  _globals['_FEATURETABLEINFO']._serialized_start=763
  _globals['_FEATURETABLEINFO']._serialized_end=844
  _globals['_FUNCTIONINFO']._serialized_start=846
  _globals['_FUNCTIONINFO']._serialized_end=878
  _globals['_FEATURESPEC']._serialized_start=881
  _globals['_FEATURESPEC']._serialized_end=1206
# @@protoc_insertion_point(module_scope)



import textwrap
import typing

from databricks.feature_store.entities.feature_functions_for_serving import (
    FeatureFunctionForServing,
    FeatureFunctionParameterInfo,
    FeatureFunctionsForServing,
)

INDENT = " " * 4


class CodegenUtils:
    """
    This utility class wraps functions used to generate the code required to evalute Feature
    Functions in Model Serving.
    """

    @staticmethod
    def _gen_params(param_infos: typing.List[FeatureFunctionParameterInfo]) -> str:
        """
        Generate a list of parameters for generated function
        """
        return ", ".join([param_info.name for param_info in param_infos])

    @staticmethod
    def _indent(body: str) -> str:
        return textwrap.indent(body, prefix=INDENT, predicate=lambda _: True)

    @staticmethod
    def generate_function(
        fn_name: str, ff_for_serving: FeatureFunctionForServing
    ) -> str:
        """
        Generates Python3 code containing a function with name `fn_name`, and parameters +
        definition as specfieid by FeatureFunctionForServing.

        For example:
        generate_function(
          "my_cool_function",
          FeatureFunctionForServing(
            ...,
            routine_definition="return a + b",
            input_params = [
              FeatureFunctionParameterInfo(name="a"),
              FeatureFunctionParameterInfo(name="b")
            ]

        =>

        '''\
        def my_cool_function(a, b):
          return a + b
        '''

        For more examples, see tests.
        """
        params_str = CodegenUtils._gen_params(ff_for_serving.input_params)
        indented_body = CodegenUtils._indent(ff_for_serving.routine_definition)
        # We don't use dedent to prettify the below since it swallows newlines.
        return f"""\
def {fn_name}({params_str}):
{indented_body}
"""



# Shared mlflow model constants that are used in both the Feature Store client and lookup client

# Module name of the original mlflow_model
MLFLOW_MODEL_NAME = "databricks.feature_store.mlflow_model"

# FeatureStoreClient.log_model will log models containing a 'raw_model' folder within the data_path.
# This folder stores the MLmodel for the raw model, which is needed to run inference.
# FeatureStoreClient.log_model may be called by any version of the FeatureStoreClient >= 0.3.0,
# however this module may be from a more recent version. As such:
# *** The value of this constant SHOULD NEVER CHANGE ***
RAW_MODEL_FOLDER = "raw_model"

# Constant for the ML model filename
ML_MODEL = "MLmodel"

# The package name of the feature lookup client as published on PyPI
FEATURE_LOOKUP_CLIENT_PIP_PACKAGE = "databricks-feature-lookup"

# The major version of the feature lookup client.  This is a shared constant in order to tightly couple:
#
# - The Feature lookup client major version (eg `{FEATURE_LOOKUP_CLIENT_PIP_PACKAGE}.0.1`)
# - The pinned major version of the databricks-feature-lookup requirement in logged model's conda.yaml
#
# If needed, these can be decoupled into separate constants, but that decision should be carefully considered.
FEATURE_LOOKUP_CLIENT_MAJOR_VERSION = 1

# Model artifact directory where we expect Feature Store internal artifacts to live. This must
# match the directory specified in ModelServingUtils:
# https://src.dev.databricks.com/databricks/universe@4dff77c752b546579f239815e520d566d2dbda20/-/blob/model-serving/model-serving-common/src/main/scala/com/databricks/modelservingcommon/utils/ModelServingUtils.scala?L117
FEATURE_STORE_INTERNAL_DATA_DIR = "_databricks_internal/"

# The default value for model params "result_type". This param is used by fe.score_batch when
# creating the spark udf.
_NO_RESULT_TYPE_PASSED = "NO_RESULT_TYPE"



""" Defines the LookupSqlEngine class, which is used to perform
lookups on SQL databases. This class differs from PublishSqlEngine in that
its actions are read-only, and uses a Python connector instead of Java.
"""

import abc
import functools
import logging
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import sqlalchemy

from databricks.feature_store.lookup_engine.lookup_engine import LookupEngine
from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.feature_store.utils.logging_utils import get_logger
from databricks.feature_store.utils.metrics_utils import (
    NAN_FEATURE_COUNT,
    LookupClientMetrics,
    lookup_call_maybe_with_metrics,
)
from databricks.feature_store.utils.sql_type_utils import (
    SQL_DATA_TYPE_CONVERTER_FACTORY,
)
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)

_logger = get_logger(__name__, log_level=logging.INFO)

LookupKeyType = Tuple[str, ...]


class LookupSqlEngine(LookupEngine):
    INFORMATION_SCHEMA = "INFORMATION_SCHEMA"
    TABLES = "TABLES"
    TABLE_CATALOG = "TABLE_CATALOG"
    TABLE_SCHEMA = "TABLE_SCHEMA"
    TABLE_NAME = "TABLE_NAME"
    COLUMNS = "COLUMNS"
    COLUMN_NAME = "COLUMN_NAME"
    COLUMN_KEY = "COLUMN_KEY"
    DATA_TYPE = "DATA_TYPE"
    TABLE_CONSTRAINTS = "TABLE_CONSTRAINTS"
    CONSTRAINT_COLUMN_USAGE = "CONSTRAINT_COLUMN_USAGE"
    CONSTRAINT_TYPE = "CONSTRAINT_TYPE"
    CONSTRAINT_NAME = "CONSTRAINT_NAME"

    @abc.abstractmethod
    def __init__(
        self, online_feature_table: OnlineFeatureTable, ro_user: str, ro_password: str
    ):
        self.online_store = online_feature_table.online_store

        result = self._get_database_and_table_name(
            online_feature_table.online_feature_table_name
        )
        (self.database_name, self.schema_name, self.table_name) = result

        self.user = ro_user
        self.password = ro_password
        self.host = self.online_store.extra_configs.host
        self.port = self.online_store.extra_configs.port

        self.primary_keys = online_feature_table.primary_keys
        self.feature_names_to_feature_details = {
            f.name: f for f in online_feature_table.features
        }
        self.features_to_type_converter = {
            feature.name: SQL_DATA_TYPE_CONVERTER_FACTORY.get_converter(feature)
            for feature in online_feature_table.features
        }
        # caching the query
        self.previous_key_columns = None
        self.previous_feature_columns = None
        self.previous_query = None

        if not self.is_lakebase_engine():
            # A validation is not really necessary. But keep it for legacy implementations
            # to make the behavior unchanged.
            self._validate_online_feature_table()

    def _get_database_and_table_name(
        self, online_table_name
    ) -> Tuple[str, Optional[str], str]:
        raise NotImplementedError

    @contextmanager
    def _get_connection(self):
        # Everything between here and the yield statement will be executed in contextmanager.__enter__()

        # We create a SQL engine instance per required online table in an inference request and dispose it afterwards.
        # A single inference request runs queries sequentially, so a connection pool is not required or beneficial.
        # TODO (ML-23188): Investigate sharing a long lived SQL engine and connection pool across inference requests.
        engine = sqlalchemy.create_engine(
            self.engine_url, poolclass=sqlalchemy.pool.NullPool
        )
        connection = engine.connect()

        # When the caller invokes "with _get_connection() as x", the connection will be returned as "x"
        yield connection

        # Everything below here will be executed in contextmanager.__exit__()

        connection.close()
        # Disposing the engine prevents connections from hanging around in memory if a connection pool is used.
        engine.dispose()

    @property
    def engine_url(self) -> str:
        raise NotImplementedError

    # Base implementation that runs a single row query for each lookup row.
    # Lakebase engine overrides this method and bulk fetches features for multiple rows.
    def _get_sql_results(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
        metrics: LookupClientMetrics = None,
    ) -> List[sqlalchemy.engine.Row]:
        if (
            self.previous_key_columns == lookup_list.columns
            and self.previous_feature_columns == feature_names
        ):
            query = self.previous_query
        else:
            pk_filter_phrase = " AND ".join(
                [f"{self._sql_safe_name(pk)} = :{pk}" for pk in lookup_list.columns]
            )
            select_feat_phrase = ", ".join(
                f"{self._sql_safe_name(f)}" for f in feature_names
            )

            safe_table_name = self._sql_safe_name(self.table_name)
            table_name_phrase = (
                f"{self._sql_safe_name(self.schema_name)}.{safe_table_name}"
                if self.schema_name
                else safe_table_name
            )

            query = sqlalchemy.sql.text(
                f"SELECT {select_feat_phrase} FROM {table_name_phrase} WHERE {pk_filter_phrase}"
            )
            self.previous_key_columns = lookup_list.columns
            self.previous_feature_columns = feature_names
            self.previous_query = query

        with self._get_connection() as sql_connection:
            run_query = functools.partial(
                self._run_lookup_sql_query,
                sql_connection,
                query,
                feature_names,
                metrics,
            )
            return [run_query(row) for row in lookup_list.items]

    # To be overridden by subclasses
    def is_lakebase_engine(self) -> bool:
        return False

    def lookup_features(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        """
        Lookups features from one table. Returns the results in format of a pandas DataFrame
        and cast the values to the exact data type as specified by FeatureLookup and FeatureFunctions.
        """
        results = self._get_sql_results(lookup_list, feature_names, metrics=metrics)

        feature_df = pd.DataFrame(results, columns=feature_names)

        feature_df.columns = feature_names
        for feature in feature_names:
            feature_df[feature] = feature_df[feature].map(
                lambda x: (
                    self.features_to_type_converter[feature].to_pandas(x) if x else x
                ),
                na_action="ignore",
            )
        return feature_df

    def lookup_feature_dicts(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
    ) -> List[Dict[str, Any]]:
        """
        Lookups features from one table. Returns the results in format of a List of Dict
        whose keys are the column names and values are the feature values.
        """
        results = self._get_sql_results(lookup_list, feature_names)
        # results cannot be empty.
        if isinstance(results[0], np.ndarray):
            # handle not found case
            return [{name: None for name in feature_names}]
        elif isinstance(results[0], dict):
            # handle case where results are already dictionaries (e.g., from lakebase engine)
            return results
        else:
            return [row._asdict() for row in results]

    def batch_lookup_features(
        self,
        lookup_list_dict: Dict[str, Dict[LookupKeyType, LookupKeyList]],
        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
        *,
        metrics: LookupClientMetrics = None,
    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
        raise NotImplementedError

    def _validate_online_feature_table(
        self,
    ) -> None:
        # Validate the Python database connection specified by the database location in OnlineFeatureTable.
        # Throw operational error if new connection is invalid
        try:
            with self._get_connection() as sql_connection:
                # Validate the online feature table exists in online database as specified by the OnlineFeatureTable.
                if not self._database_contains_feature_table(sql_connection):
                    raise ValueError(
                        f"Table {self.table_name} does not exist in database {self.database_name}."
                    )

                    # Validate the online feature table has the same primary keys specified by the OnlineFeatureTable.
                if not self._database_contains_primary_keys(sql_connection):
                    raise ValueError(
                        f"Table {self.table_name} does not contain primary keys {self.primary_keys}."
                    )
        except Exception as e:
            raise ValueError(f"Connection could not be established: {str(e)}.")

    def shutdown(self) -> None:
        """
        Closes the database connection if it exists on the SQL lookup engine.
        :return:
        """
        self._close()

    def _close(self) -> None:
        """
        This is a no-op because a new sql connection is opened for each query and closed after the query executes.
        """
        pass

    def _run_lookup_sql_query(
        self,
        sql_connection,
        query,
        feature_names,
        metrics,
        lookup_row: List[Tuple[str, Any]],
    ):
        """
        This helper function executes a single SQL query .
        """
        query_params = dict(lookup_row)
        run_lookup = lookup_call_maybe_with_metrics(sql_connection.execute, metrics)
        sql_data = run_lookup(query, query_params)
        feat_values = sql_data.fetchall()
        if len(feat_values) == 0:
            _logger.warning(
                f"No feature values found in {self.table_name} for {query_params}."
            )
            nan_features = np.empty(len(feature_names))
            nan_features[:] = np.nan
            if metrics:
                # 0 nan feature count still needs to be recorded to propogate metrics for inference request
                metrics.increase_metric(NAN_FEATURE_COUNT, len(nan_features))

            return nan_features
        # Return the first result
        return feat_values[0]

    @classmethod
    def _sql_safe_name(cls, name):
        raise NotImplementedError

    def _database_contains_feature_table(self, sql_connection):
        raise NotImplementedError

    def _database_contains_primary_keys(self, sql_connection):
        raise NotImplementedError



# This file contains all the environment variables used by the lookup client.


# The provisioner of this model is expected to set an environment variable with the path to a
# feature_tables_for_serving.dat file.
# - Set by serving scheduler
FEATURE_TABLES_FOR_SERVING_FILEPATH_ENV = "FEATURE_TABLES_FOR_SERVING_FILEPATH"
# should match LOOKUP_CLIENT_FEATURE_FUNCTION_EVALUATION_ENABLED_ENV in
# model-serving/model-serving-common/src/main/scala/com/databricks/mlflow/utils/serving/FeatureStoreConstants.scala
# - Set by serving scheduler
LOOKUP_CLIENT_FEATURE_FUNCTION_EVALUATION_ENABLED_ENV = (
    "LOOKUP_CLIENT_FEATURE_FUNCTION_EVALUATION_ENABLED"
)
# Enable metrics recorder for feature lookup.
# should match ENABLE_FEATURE_LOOKUP_METRICS_RECORDER_ENV in
# model-serving/model-serving-common/src/main/scala/com/databricks/mlflow/utils/serving/FeatureStoreConstants.scala
# - Set by serving scheduler
ENABLE_FEATURE_LOOKUP_METRICS_RECORDER_ENV = "ENABLE_FEATURE_LOOKUP_METRICS_RECORDER"
# Enable debug logging. The lookup logic will print info logs in service logs with a break down of the latency numbers
# - Set by customers through environment variables config.
LOOKUP_PERFORMANCE_DEBUG_ENABLED = "LOOKUP_PERFORMANCE_DEBUG_ENABLED"
# Env var to tell if the served entity is a Model or a FeatureSpec
# - Set by serving scheduler
ENV_SERVABLE_TYPE = "SERVABLE_TYPE"
# If set to true, postgres connection is allowed.
# Note that the actual eligibility is determined by the features used.
# - Set by customers through environment variables config.
FEATURE_SERVING_HP = "FEATURE_SERVING_HP"
# FEATURE_SERVING_CONNECTION_POOL_SIZE overrides the default connection pool size
# calculated by the utils in lakebase_utils.py, giving the backend or users the
# opportunity to tune the connection pool size for their use cases.
# - Set by serving scheduler but not used yet in Dec 2025.
FEATURE_SERVING_CONNECTION_POOL_SIZE = "FEATURE_SERVING_CONNECTION_POOL_SIZE"
# Engine configurations
# - Set by serving scheduler but not used yet in Dec 2025.
DISABLE_CONNECTION_POOL = "DISABLE_CONNECTION_POOL"

######## Dynamic environment variables #########
# Some secrets related environment variables names are constructed dynamically in
# feature-store/lookup-client/python/databricks/feature_store/online_lookup_client.py
# with certain suffixes. For example:
# _USER
# _PASSWORD
# _ACCESS_KEY_ID
# _SECRET_ACCESS_KEY
# _AUTHORIZATION_KEY

####### Deprecated environment variables ########

# The following environment variable names are deprecated in previouse versions
# Do not use them in new code because it may conflict with old version lookup
# clients that are still running in production.
# Note: they are env var names, not Python constant variable names.

# 1. LOOKUP_CLIENT_INSTRUMENTATION_ENABLED
# 2. ENABLE_LOOKUP_CLIENT_INSTRUMENTATION



""" Defines the type conversion classes from pandas to cosmosdb and vice versa.
"""

from databricks.feature_store.utils.pandas_type_utils import (
    PandasArrayTypeConverter,
    PandasBase64BinaryTypeConverter,
    PandasBooleanTypeConverter,
    PandasDecimalTypeConverter,
    PandasDoubleTypeConverter,
    PandasFloatTypeConverter,
    PandasIntTypeConverter,
    PandasIsoDateTypeConverter,
    PandasIsoTimestampTypeUsConverter,
    PandasLongTypeConverter,
    PandasMapTypeConverter,
    PandasShortTypeConverter,
    PandasStringTypeConverter,
)
from databricks.ml_features_common.entities.data_type import DataType
from databricks.ml_features_common.utils.converter_utils import ConverterFactory

"""
All supported converters.
See the link below for relevant design decision on Cosmos DB type conversion.
https://docs.google.com/document/d/1cG7PDWHld-WwD2UWp7v80KJT7lXOvvO0wNyKuZG2VQ8/edit#heading=h.24cgj83368tv
"""
BASIC_DATA_TYPE_CONVERTERS = {
    DataType.SHORT: PandasShortTypeConverter,
    DataType.INTEGER: PandasIntTypeConverter,
    DataType.LONG: PandasLongTypeConverter,
    DataType.FLOAT: PandasFloatTypeConverter,
    DataType.DOUBLE: PandasDoubleTypeConverter,
    DataType.BOOLEAN: PandasBooleanTypeConverter,
    DataType.STRING: PandasStringTypeConverter,
    DataType.TIMESTAMP: PandasIsoTimestampTypeUsConverter,
    DataType.DATE: PandasIsoDateTypeConverter,
    DataType.BINARY: PandasBase64BinaryTypeConverter,
}

COMPLEX_DATA_TYPE_CONVERTERS = {
    DataType.DECIMAL: PandasDecimalTypeConverter,
    DataType.ARRAY: PandasArrayTypeConverter,
    DataType.MAP: PandasMapTypeConverter,
}

COSMOSDB_DATA_TYPE_CONVERTER_FACTORY = ConverterFactory(
    BASIC_DATA_TYPE_CONVERTERS, COMPLEX_DATA_TYPE_CONVERTERS
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: feature_store_serving.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    4,
    '',
    'feature_store_serving.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x65\x61ture_store_serving.proto\x12\x12\x66\x65\x61turestorecommon\x1a\x15scalapb/scalapb.proto\"\x9b\x04\n\x15OnlineStoreForServing\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x1a\n\x12read_secret_prefix\x18\x03 \x01(\t\x12\x1d\n\x15\x63reation_timestamp_ms\x18\x04 \x01(\x03\x12\x33\n\nmysql_conf\x18\x05 \x01(\x0b\x32\x1d.featurestorecommon.MySqlConfH\x00\x12<\n\x0fsql_server_conf\x18\x06 \x01(\x0b\x32!.featurestorecommon.SqlServerConfH\x00\x12\x39\n\rdynamodb_conf\x18\x07 \x01(\x0b\x32 .featurestorecommon.DynamoDbConfH\x00\x12\x39\n\rcosmosdb_conf\x18\x08 \x01(\x0b\x32 .featurestorecommon.CosmosDbConfH\x00\x12=\n\x0f\x62rickstore_conf\x18\t \x01(\x0b\x32\".featurestorecommon.BrickstoreConfH\x00\x12\x31\n\nquery_mode\x18\x15 \x01(\x0e\x32\x1d.featurestorecommon.QueryModeB\x0f\n\rextra_configs\"\'\n\tMySqlConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"+\n\rSqlServerConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"\x1e\n\x0c\x44ynamoDbConf\x12\x0e\n\x06region\x18\x01 \x01(\t\"#\n\x0c\x43osmosDbConf\x12\x13\n\x0b\x61\x63\x63ount_uri\x18\x01 \x01(\t\"^\n\x0e\x42rickstoreConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\x12\x19\n\x11table_serving_url\x18\x03 \x01(\t\x12\x15\n\rinstance_name\x18\x04 \x01(\t\"j\n\x0e\x46\x65\x61tureDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\x12\x19\n\x11\x64\x61ta_type_details\x18\x03 \x01(\t\"R\n\x11PrimaryKeyDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"T\n\x13TimestampKeyDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"\xe2\x02\n\x12OnlineFeatureTable\x12\x1a\n\x12\x66\x65\x61ture_table_name\x18\x01 \x01(\t\x12!\n\x19online_feature_table_name\x18\x02 \x01(\t\x12?\n\x0conline_store\x18\x03 \x01(\x0b\x32).featurestorecommon.OnlineStoreForServing\x12;\n\x0cprimary_keys\x18\x04 \x03(\x0b\x32%.featurestorecommon.PrimaryKeyDetails\x12\x18\n\x10\x66\x65\x61ture_table_id\x18\x05 \x01(\t\x12\x34\n\x08\x66\x65\x61tures\x18\x06 \x03(\x0b\x32\".featurestorecommon.FeatureDetails\x12?\n\x0etimestamp_keys\x18\x07 \x03(\x0b\x32\'.featurestorecommon.TimestampKeyDetails\"\xa4\x01\n\x17\x46\x65\x61tureTablesForServing\x12=\n\ronline_tables\x18\x01 \x03(\x0b\x32&.featurestorecommon.OnlineFeatureTable\x12J\n\x17\x66\x65\x61ture_definition_type\x18\x02 \x01(\x0e\x32).featurestorecommon.FeatureDefinitionType\"+\n\rMySqlMetadata\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"/\n\x11SqlServerMetadata\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"B\n\x10\x44ynamoDbMetadata\x12\x0e\n\x06region\x18\x01 \x01(\t\x12\x11\n\ttable_arn\x18\x02 \x01(\t\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\"K\n\x10\x43osmosDbMetadata\x12\x13\n\x0b\x61\x63\x63ount_uri\x18\x01 \x01(\t\x12\x15\n\rcontainer_uri\x18\x02 \x01(\t\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\"\xa9\x04\n\x0bOnlineStore\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x04 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x05 \x01(\x03\x12\x12\n\ncreator_id\x18\x06 \x01(\t\x12\x10\n\x04host\x18\x07 \x01(\tB\x02\x18\x01\x12\x10\n\x04port\x18\x08 \x01(\x05\x42\x02\x18\x01\x12\x1b\n\x13last_update_user_id\x18\t \x01(\t\x12;\n\x0emysql_metadata\x18\n \x01(\x0b\x32!.featurestorecommon.MySqlMetadataH\x00\x12\x44\n\x13sql_server_metadata\x18\x0b \x01(\x0b\x32%.featurestorecommon.SqlServerMetadataH\x00\x12\x41\n\x11\x64ynamodb_metadata\x18\x0c \x01(\x0b\x32$.featurestorecommon.DynamoDbMetadataH\x00\x12\x41\n\x11\x63osmosdb_metadata\x18\r \x01(\x0b\x32$.featurestorecommon.CosmosDbMetadataH\x00\x42\x15\n\x13\x61\x64\x64itional_metadata\"\xdf\x04\n\x13OnlineStoreDetailed\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x04 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x05 \x01(\x03\x12\x12\n\ncreator_id\x18\x06 \x01(\t\x12\x10\n\x04host\x18\x07 \x01(\tB\x02\x18\x01\x12\x10\n\x04port\x18\x08 \x01(\x05\x42\x02\x18\x01\x12\x1b\n\x13last_update_user_id\x18\t \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\n \x03(\t\x12\x1a\n\x12read_secret_prefix\x18\x0b \x01(\t\x12;\n\x0emysql_metadata\x18\x0c \x01(\x0b\x32!.featurestorecommon.MySqlMetadataH\x00\x12\x44\n\x13sql_server_metadata\x18\r \x01(\x0b\x32%.featurestorecommon.SqlServerMetadataH\x00\x12\x41\n\x11\x64ynamodb_metadata\x18\x0e \x01(\x0b\x32$.featurestorecommon.DynamoDbMetadataH\x00\x12\x41\n\x11\x63osmosdb_metadata\x18\x0f \x01(\x0b\x32$.featurestorecommon.CosmosDbMetadataH\x00\x42\x15\n\x13\x61\x64\x64itional_metadata\"\x83\x01\n\x1c\x46\x65\x61tureFunctionParameterInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\ttype_text\x18\x02 \x01(\t\x12\x11\n\ttype_json\x18\x03 \x01(\t\x12/\n\ttype_name\x18\x04 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"\xdb\x01\n\x19\x46\x65\x61tureFunctionForServing\x12\x11\n\tfull_name\x18\x01 \x01(\t\x12\x46\n\x0cinput_params\x18\x02 \x03(\x0b\x32\x30.featurestorecommon.FeatureFunctionParameterInfo\x12/\n\tdata_type\x18\x03 \x01(\x0e\x32\x1c.featurestorecommon.DataType\x12\x16\n\x0e\x66ull_data_type\x18\x04 \x01(\t\x12\x1a\n\x12routine_definition\x18\x05 \x01(\t\"c\n\x1a\x46\x65\x61tureFunctionsForServing\x12\x45\n\x0e\x66\x66_for_serving\x18\x01 \x03(\x0b\x32-.featurestorecommon.FeatureFunctionForServing*$\n\x05\x43loud\x12\x07\n\x03\x41WS\x10\x01\x12\t\n\x05\x41ZURE\x10\x02\x12\x07\n\x03GCP\x10\x03*\x81\x01\n\tStoreType\x12\x10\n\x0c\x41URORA_MYSQL\x10\x01\x12\x0e\n\nSQL_SERVER\x10\x02\x12\t\n\x05MYSQL\x10\x03\x12\x0c\n\x08\x44YNAMODB\x10\x04\x12\x0c\n\x08\x43OSMOSDB\x10\x05\x12\x0e\n\nBRICKSTORE\x10\x06\x12\x1b\n\x17\x44\x41TABRICKS_ONLINE_STORE\x10\x07*\xae\x01\n\x08\x44\x61taType\x12\x0b\n\x07INTEGER\x10\x01\x12\t\n\x05\x46LOAT\x10\x02\x12\x0b\n\x07\x42OOLEAN\x10\x03\x12\n\n\x06STRING\x10\x04\x12\n\n\x06\x44OUBLE\x10\x05\x12\x08\n\x04LONG\x10\x06\x12\r\n\tTIMESTAMP\x10\x07\x12\x08\n\x04\x44\x41TE\x10\x08\x12\t\n\x05SHORT\x10\t\x12\t\n\x05\x41RRAY\x10\n\x12\x07\n\x03MAP\x10\x0b\x12\n\n\x06\x42INARY\x10\x0c\x12\x0b\n\x07\x44\x45\x43IMAL\x10\r\x12\n\n\x06STRUCT\x10\x0e*4\n\tQueryMode\x12\x16\n\x12PRIMARY_KEY_LOOKUP\x10\x00\x12\x0f\n\x0bRANGE_QUERY\x10\x01*m\n\x15\x46\x65\x61tureDefinitionType\x12\'\n#FEATURE_DEFINITION_TYPE_UNSPECIFIED\x10\x00\x12\x11\n\rFEATURE_TABLE\x10\x01\x12\x18\n\x14\x44\x45\x43LARATIVE_FEATURES\x10\x02\x42}\n.com.databricks.proto.featurestorecommon.jprotoB\x18\x46\x65\x61tureStoreServingProtoP\x01\xa0\x01\x01\xe2?+\n\'com.databricks.proto.featurestorecommon\x10\x01')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feature_store_serving_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n.com.databricks.proto.featurestorecommon.jprotoB\030FeatureStoreServingProtoP\001\240\001\001\342?+\n\'com.databricks.proto.featurestorecommon\020\001'
  _globals['_ONLINESTORE'].fields_by_name['host']._loaded_options = None
  _globals['_ONLINESTORE'].fields_by_name['host']._serialized_options = b'\030\001'
  _globals['_ONLINESTORE'].fields_by_name['port']._loaded_options = None
  _globals['_ONLINESTORE'].fields_by_name['port']._serialized_options = b'\030\001'
  _globals['_ONLINESTOREDETAILED'].fields_by_name['host']._loaded_options = None
  _globals['_ONLINESTOREDETAILED'].fields_by_name['host']._serialized_options = b'\030\001'
  _globals['_ONLINESTOREDETAILED'].fields_by_name['port']._loaded_options = None
  _globals['_ONLINESTOREDETAILED'].fields_by_name['port']._serialized_options = b'\030\001'
  _globals['_CLOUD']._serialized_start=3531
  _globals['_CLOUD']._serialized_end=3567
  _globals['_STORETYPE']._serialized_start=3570
  _globals['_STORETYPE']._serialized_end=3699
  _globals['_DATATYPE']._serialized_start=3702
  _globals['_DATATYPE']._serialized_end=3876
  _globals['_QUERYMODE']._serialized_start=3878
  _globals['_QUERYMODE']._serialized_end=3930
  _globals['_FEATUREDEFINITIONTYPE']._serialized_start=3932
  _globals['_FEATUREDEFINITIONTYPE']._serialized_end=4041
  _globals['_ONLINESTOREFORSERVING']._serialized_start=75
  _globals['_ONLINESTOREFORSERVING']._serialized_end=614
  _globals['_MYSQLCONF']._serialized_start=616
  _globals['_MYSQLCONF']._serialized_end=655
  _globals['_SQLSERVERCONF']._serialized_start=657
  _globals['_SQLSERVERCONF']._serialized_end=700
  _globals['_DYNAMODBCONF']._serialized_start=702
  _globals['_DYNAMODBCONF']._serialized_end=732
  _globals['_COSMOSDBCONF']._serialized_start=734
  _globals['_COSMOSDBCONF']._serialized_end=769
  _globals['_BRICKSTORECONF']._serialized_start=771
  _globals['_BRICKSTORECONF']._serialized_end=865
  _globals['_FEATUREDETAILS']._serialized_start=867
  _globals['_FEATUREDETAILS']._serialized_end=973
  _globals['_PRIMARYKEYDETAILS']._serialized_start=975
  _globals['_PRIMARYKEYDETAILS']._serialized_end=1057
  _globals['_TIMESTAMPKEYDETAILS']._serialized_start=1059
  _globals['_TIMESTAMPKEYDETAILS']._serialized_end=1143
  _globals['_ONLINEFEATURETABLE']._serialized_start=1146
  _globals['_ONLINEFEATURETABLE']._serialized_end=1500
  _globals['_FEATURETABLESFORSERVING']._serialized_start=1503
  _globals['_FEATURETABLESFORSERVING']._serialized_end=1667
  _globals['_MYSQLMETADATA']._serialized_start=1669
  _globals['_MYSQLMETADATA']._serialized_end=1712
  _globals['_SQLSERVERMETADATA']._serialized_start=1714
  _globals['_SQLSERVERMETADATA']._serialized_end=1761
  _globals['_DYNAMODBMETADATA']._serialized_start=1763
  _globals['_DYNAMODBMETADATA']._serialized_end=1829
  _globals['_COSMOSDBMETADATA']._serialized_start=1831
  _globals['_COSMOSDBMETADATA']._serialized_end=1906
  _globals['_ONLINESTORE']._serialized_start=1909
  _globals['_ONLINESTORE']._serialized_end=2462
  _globals['_ONLINESTOREDETAILED']._serialized_start=2465
  _globals['_ONLINESTOREDETAILED']._serialized_end=3072
  _globals['_FEATUREFUNCTIONPARAMETERINFO']._serialized_start=3075
  _globals['_FEATUREFUNCTIONPARAMETERINFO']._serialized_end=3206
  _globals['_FEATUREFUNCTIONFORSERVING']._serialized_start=3209
  _globals['_FEATUREFUNCTIONFORSERVING']._serialized_end=3428
  _globals['_FEATUREFUNCTIONSFORSERVING']._serialized_start=3430
  _globals['_FEATUREFUNCTIONSFORSERVING']._serialized_end=3529
# @@protoc_insertion_point(module_scope)



from collections import defaultdict
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import requests
from requests.adapters import HTTPAdapter

from databricks.feature_store.lookup_engine.lookup_engine import LookupEngine
from databricks.feature_store.lookup_engine.oauth_token_manager import OAuthTokenManager
from databricks.feature_store.utils.brickstore_http_type_utils import (
    BRICKSTORE_HTTP_DATA_TYPE_CONVERTER_FACTORY,
)
from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.feature_store.utils.lakebase_constants import (
    BRICKSTORE_OAUTH_TOKEN_FILE_PATH,
)
from databricks.feature_store.utils.metrics_utils import (
    LookupClientMetrics,
    lookup_call_maybe_with_metrics,
)
from databricks.feature_store.utils.retry_utils import RetryWithLogging
from databricks.feature_store.utils.trace_utils import trace_latency
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)

LookupKeyType = Tuple[str, ...]


class LookupLegacyBrickstoreHttpEngine(LookupEngine):
    """
    Read online features from Brickstore by making requests aginst the HTTP gateway.
    """

    def __init__(
        self,
        online_feature_tables: Union[OnlineFeatureTable, List[OnlineFeatureTable]],
        disable_connection_pool: bool = False,
    ):
        # Table serving URL will be the same for all tables, so we can read the first.
        # Consistency is enforced by the control plane.
        if isinstance(online_feature_tables, OnlineFeatureTable):
            online_feature_tables = [online_feature_tables]

        self._table_serving_url = online_feature_tables[
            0
        ].online_store.extra_configs.table_serving_url

        if not disable_connection_pool:
            self._connection = self._new_session()
        else:
            self._connection = requests

        self._oauth_token_manager = OAuthTokenManager(
            oauth_token_file_path=BRICKSTORE_OAUTH_TOKEN_FILE_PATH,
        )
        self._oauth_token_manager.start_token_refresh_thread()

        self._columns_to_type_converter_map = {
            oft.online_feature_table_name: {
                col.name: BRICKSTORE_HTTP_DATA_TYPE_CONVERTER_FACTORY.get_converter(col)
                for col in oft.primary_keys + oft.features
            }
            for oft in online_feature_tables
        }

    @staticmethod
    def _new_session() -> requests.Session:
        session = requests.Session()
        # Retry on the given list of status and connection errors e.g. RemoteDisconnected.
        # When the retry attempt is exhausted, if the last error was a status code, the response
        # will be returned. Otherwise, if the last error was an exception, it will be raised.
        retry = RetryWithLogging(
            total=3,
            backoff_factor=0.2,
            status_forcelist=[429, 500, 502, 503, 504],
            # Default methods to be used for allowed_methods
            # We are adding POST into the method
            # DEFAULT_ALLOWED_METHODS = frozenset({'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PUT', 'TRACE'})
            # See https://urllib3.readthedocs.io/en/stable/reference/urllib3.util.html#urllib3.util.Retry.DEFAULT_ALLOWED_METHODS
            allowed_methods=frozenset(
                ["DELETE", "GET", "HEAD", "OPTIONS", "PUT", "TRACE", "POST"]
            ),
            raise_on_status=False,
        )
        session.mount("https://", HTTPAdapter(max_retries=retry))
        return session

    def _pandas_to_json_input_dict(
        self, feature_table_name: str, row: List[Tuple[str, Any]]
    ) -> Dict[str, Any]:
        """
        Converts the input Pandas row to JSON inputs
        """
        return {
            col_name: self._columns_to_type_converter_map[feature_table_name][
                col_name
            ].to_online_store(col_value)
            for col_name, col_value in row
        }

    def _raw_response_to_typed(
        self, feature_table_name: str, feature_name: str, feature_value: Any
    ) -> Any:
        """
        Converts the input result with brickstore-returned python values to pandas types based on
        the input features_names and features converter.
        """
        return self._columns_to_type_converter_map[feature_table_name][
            feature_name
        ].to_pandas(feature_value)

    def _lookup_list_dict_to_request_json(
        self,
        lookup_list_dict: Dict[str, Dict[LookupKeyType, LookupKeyList]],
        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
    ):
        tables_dict = {}

        for oft_name, lookup_to_pk_values_dict in lookup_list_dict.items():
            selected_features_for_table = set()
            pk_input_values_for_table = []

            def convert_fn(row):
                return self._pandas_to_json_input_dict(oft_name, row)

            for lookup_key, pk_list in lookup_to_pk_values_dict.items():
                features_for_this_lookup_key = feature_names_dict[oft_name][lookup_key]
                # Always want to look up PKs.
                selected_features_for_table.update(list(pk_list.columns))
                selected_features_for_table.update(features_for_this_lookup_key)

                lookup_dict = [convert_fn(row) for row in pk_list.items]
                pk_input_values_for_table += lookup_dict

            table_dict = {
                "keys": pk_input_values_for_table,
                "select": list(selected_features_for_table),
            }

            tables_dict[oft_name] = table_dict
        return tables_dict

    def _response_to_looked_up_df_dict(
        self,
        lookup_list_dict: Dict[str, Dict[LookupKeyType, LookupKeyList]],
        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
        results: Dict,
    ):
        result_dict = defaultdict(lambda: defaultdict(list))

        for oft_name, table_results_dict in results.items():
            if "rows" not in table_results_dict:
                table_results_dict["rows"] = []
            looked_up_data = table_results_dict["rows"]
            col_infos = table_results_dict["schema"]["columns"]

            def convert_fn(feature_name, feature_value):
                return self._raw_response_to_typed(
                    oft_name, feature_name, feature_value
                )

            looked_up_data_columns = [col_info["name"] for col_info in col_infos]
            converted_looked_up_data_dict = {
                looked_up_data_columns[i]: [
                    convert_fn(looked_up_data_columns[i], row[i])
                    for row in looked_up_data
                ]
                for i in range(len(looked_up_data_columns))
            }
            # Example:
            # converted_looked_up_data_dict = {
            #     "a": [1, 3, 5, 7],
            #     "b": [2, 4, 6, 8],
            #     "name": ["abc", "def", "ghi", "jkl"],
            #     "age": [10, 20, 30, 40],
            # }

            lookup_key_to_input_pks_list_dict = lookup_list_dict[oft_name]
            for (
                lookup_key,
                input_pks_list,
            ) in lookup_key_to_input_pks_list_dict.items():
                # Example:
                # input_pks_list = LookupKeyList(columns=["a", "b"], rows=[(1, 2), (3, 4)])

                feat_names_to_look_up = feature_names_dict[oft_name][lookup_key]
                # Example:
                # feat_names_to_look_up = ["name", "age"]

                # Input pks df is raw input from user, so we need to convert it to same types
                # as converted looked up df for merge to work.
                # Compute the converted primary key values in user's request.
                converted_input_pks = [
                    tuple(convert_fn(pk_name, pk_value) for pk_name, pk_value in row)
                    for row in input_pks_list.items
                ]
                # Example: converted_input_pks = [(1, 2), (3, 4)]

                # Compute the converted primary key values in the looked up data.
                converted_looked_up_data_pks = zip(
                    *[
                        converted_looked_up_data_dict[pk_name]
                        for pk_name in input_pks_list.columns
                    ]
                )
                # Example: converted_looked_up_data_pks = [(1, 2), (3, 4), (5, 6), (7, 8)]

                # Construct a dict that maps the converted primary key values in the looked up data
                # to its row index.
                converted_looked_up_data_pks_to_row_index = {
                    pk: i for i, pk in enumerate(converted_looked_up_data_pks)
                }
                # Example:
                # converted_looked_up_data_pks_to_row_index = {
                #     (1, 2): 0,
                #     (3, 4): 1,
                #     (5, 6): 2,
                #     (7, 8): 3,
                # }

                # Produce the final results.
                # For each row, get the looked up values for the requested features.
                def get_result_value(row_index, feat_name):
                    column_values = converted_looked_up_data_dict.get(feat_name, None)
                    if column_values is None:
                        return np.nan

                    return column_values[row_index]

                def get_result_row(input_pk):
                    row_index = converted_looked_up_data_pks_to_row_index.get(
                        input_pk, None
                    )
                    if row_index is None:
                        return np.full(len(feat_names_to_look_up), np.nan)

                    return [
                        get_result_value(row_index, feat_name)
                        for feat_name in feat_names_to_look_up
                    ]

                converted_results = [
                    get_result_row(input_pk) for input_pk in converted_input_pks
                ]
                result_dict[oft_name][lookup_key] = pd.DataFrame(
                    converted_results, columns=feat_names_to_look_up
                )

        return {k: dict(v) for k, v in result_dict.items()}

    def _lookup_and_aggregate_features(self, requested_table_info_dict):
        """
        Send lookup request. If the response contains a page token,
        send requests until there is no page token and aggregate responses.
        """

        @trace_latency("online_feature_store")
        def _send_req(page_token=None, is_retry=False):
            oauth_token = self._oauth_token_manager.get_oauth_token_or_password()
            headers = {"Authorization": f"Bearer {oauth_token}".strip()}
            req_json = {
                "tables": requested_table_info_dict,
            }
            if page_token is not None:
                req_json["page_token"] = page_token
            response = self._connection.post(
                self._table_serving_url, json=req_json, headers=headers
            )
            if response.status_code == 200:
                return response.json()
            elif response.status_code == 401:
                if is_retry:
                    raise Exception(f"Expired token: {response.text}")
                else:
                    self._oauth_token_manager.refresh_oauth_token()
                    return _send_req(page_token, is_retry=True)
            elif (
                response.status_code == 400
                and "Server received a request which exceeds maximum allowed content length"
                in response.text
            ):
                raise Exception(
                    f"Request size too large. Please reduce the number of requested rows: {response.text}"
                )
            else:
                raise Exception(f"status_code {response.status_code}: {response.text}")

        resp = _send_req()
        if "results" in resp:
            aggregated_table_info = resp["results"]
        else:
            raise Exception(
                "No online feature information found. Ensure all feature tables are synchronized to an online store."
            )
        while "next_page_token" in resp:
            resp = _send_req(resp["next_page_token"])
            for table, table_info in resp["results"].items():
                if table not in aggregated_table_info:
                    aggregated_table_info[table] = resp["results"][table]
                else:
                    aggregated_table_info[table]["rows"] += table_info["rows"]

        return aggregated_table_info

    def batch_lookup_features(
        self,
        lookup_list_dict: Dict[str, Dict[LookupKeyType, LookupKeyList]],
        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
        *,
        metrics: LookupClientMetrics = None,
    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
        """
        Assume the following lookup_list_dict:
        { table_a: {
          ("my_name",): LookupKeyList(columns=["name_pk"], rows=["x1", "x2"]),
          ("friend_name",): LookupKeyList(columns=["name_pk"], rows=["y1", "y2"])
        }}

        We need to turn into this request JSON:
        {
            "tables": {
                "catalog.schema.table_a": {
                    "keys": [
                        {"name_pk": "x1"},
                        {"name_pk": "y1"},
                        {"name_pk": "x2"},
                        {"name_pk": "y2"}
                    ],
                    "select": ["phone_number", "address"]
                }
            }
        }

        Assuming this provides the response:
        {
            "results": {
                "catalog.schema.table_a": {
                    "schema": {
                        "columns": [
                            {"name": "name", "type_name": "STRING", "nullable": false},
                            {"name": "phone_number", "type_name": "STRING", "nullable": false},
                            {"name": "address", "type_name": "STRING", "nullable": true}
                        ]
                    },
                    "rows": [
                        ["x1", "123", "abc"],
                        ["y1", "456", "def"],
                        ["x2", "789", "ghi"],
                        ["y2", "000", "jkl"],
                    ]
                },
            },
            "next_page_token": "Y2F0YWxvZy5zY2hlbWEudGFibGUyOjE="
        }

        We then need to turn into this map:
        {
            "table_a": {
                ("my_name",): pd.DataFrame({"phone_number": [123, 789]),
                ("friend_name",): pd.DataFrame({"address": [def, jkl})
            }
        }

        """
        tables_info = self._lookup_list_dict_to_request_json(
            lookup_list_dict, feature_names_dict
        )
        lookup_and_aggregate_features = lookup_call_maybe_with_metrics(
            self._lookup_and_aggregate_features, metrics
        )
        looked_up_features = lookup_and_aggregate_features(tables_info)
        return self._response_to_looked_up_df_dict(
            lookup_list_dict, feature_names_dict, looked_up_features
        )

    def lookup_features(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        raise NotImplementedError

    def shutdown(self) -> None:
        self._run_token_refresh_thread = False



from typing import Union

from databricks.ml_features_common.entities._feature_store_object import (
    _FeatureStoreObject,
)
from databricks.ml_features_common.entities.cloud import Cloud
from databricks.ml_features_common.entities.query_mode import QueryMode
from databricks.ml_features_common.entities.store_type import StoreType


class BrickstoreConf(_FeatureStoreObject):
    def __init__(self, host: str, port: int, table_serving_url: str):
        self._host = host
        self._port = port
        self._table_serving_url = table_serving_url

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def table_serving_url(self):
        return self._table_serving_url


class MySqlConf(_FeatureStoreObject):
    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port


class SqlServerConf(_FeatureStoreObject):
    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port


class DynamoDbConf(_FeatureStoreObject):
    def __init__(self, region: str):
        self._region = region

    @property
    def region(self):
        return self._region


class CosmosDbConf(_FeatureStoreObject):
    def __init__(self, account_uri: str):
        self._account_uri = account_uri

    @property
    def account_uri(self):
        return self._account_uri


class OnlineStoreForServing:
    def __init__(
        self,
        cloud: Cloud,
        store_type: StoreType,
        read_secret_prefix: str,
        creation_timestamp_ms: int,
        extra_configs: Union[BrickstoreConf, MySqlConf, SqlServerConf, DynamoDbConf],
        query_mode: QueryMode,
    ):
        self._cloud = cloud
        self._store_type = store_type
        self._read_secret_prefix = read_secret_prefix
        self._creation_timestamp_ms = creation_timestamp_ms
        self._extra_configs = extra_configs
        self._query_mode = query_mode

    @property
    def extra_configs(self):
        return self._extra_configs

    @property
    def creation_timestamp_ms(self):
        return self._creation_timestamp_ms

    @property
    def query_mode(self) -> QueryMode:
        return self._query_mode

    @property
    def cloud(self):
        return self._cloud

    @property
    def store_type(self):
        return self._store_type

    @property
    def read_secret_prefix(self):
        return self._read_secret_prefix

    @classmethod
    def from_proto(cls, the_proto):
        conf = None
        if the_proto.WhichOneof("extra_configs") == "mysql_conf":
            conf = MySqlConf(the_proto.mysql_conf.host, the_proto.mysql_conf.port)
        elif the_proto.WhichOneof("extra_configs") == "sql_server_conf":
            conf = SqlServerConf(
                the_proto.sql_server_conf.host, the_proto.sql_server_conf.port
            )
        elif the_proto.WhichOneof("extra_configs") == "brickstore_conf":
            conf = BrickstoreConf(
                the_proto.brickstore_conf.host,
                the_proto.brickstore_conf.port,
                the_proto.brickstore_conf.table_serving_url,
            )
        elif the_proto.WhichOneof("extra_configs") == "dynamodb_conf":
            conf = DynamoDbConf(the_proto.dynamodb_conf.region)
        elif the_proto.WhichOneof("extra_configs") == "cosmosdb_conf":
            conf = CosmosDbConf(the_proto.cosmosdb_conf.account_uri)
        else:
            raise ValueError("Unsupported Store Type: " + str(the_proto))

        # The `query_mode` should always be defined. We explicitly check this to avoid erroneously retrieving the
        # default QueryMode enum value `PRIMARY_KEY_LOOKUP` when the `query_mode` proto field is not set.
        if not the_proto.HasField("query_mode"):
            raise ValueError(f"'query_mode' should be defined: {the_proto}")
        return cls(
            cloud=the_proto.cloud,
            store_type=the_proto.store_type,
            extra_configs=conf,
            read_secret_prefix=the_proto.read_secret_prefix,
            creation_timestamp_ms=the_proto.creation_timestamp_ms,
            query_mode=the_proto.query_mode,
        )



import os
from typing import List

from databricks.ml_features_common.entities._feature_store_object import (
    _FeatureStoreObject,
)
from databricks.ml_features_common.entities.feature_definition_type import (
    FeatureDefinitionType,
)
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    FeatureTablesForServing as ProtoFeatureTablesForServing,
)


class FeatureTablesForServing(_FeatureStoreObject):
    DATA_FILE = "feature_tables_for_serving.dat"

    def __init__(
        self,
        online_feature_tables: List[OnlineFeatureTable],
        feature_definition_type: FeatureDefinitionType,
    ):
        self._online_feature_tables = online_feature_tables
        self._feature_definition_type = feature_definition_type

    @property
    def online_feature_tables(self) -> List[OnlineFeatureTable]:
        return self._online_feature_tables

    @property
    def feature_definition_type(self) -> FeatureDefinitionType:
        """
        Returns the feature definition type indicating how features are defined.

        For DECLARATIVE_FEATURES, FeatureTablesForServing is the only source of information for the lookup client.
        For FEATURE_TABLE, the lookup client will depend on both FeatureTablesForServing and the model's FeatureSpec.
        """
        return self._feature_definition_type

    @classmethod
    def from_proto(cls, the_proto: ProtoFeatureTablesForServing):
        online_fts = [
            OnlineFeatureTable.from_proto(online_table)
            for online_table in the_proto.online_tables
        ]

        # Default to FEATURE_TABLE if the feature_definition_type is not set
        feature_definition_type = FeatureDefinitionType.FEATURE_TABLE
        if the_proto.HasField("feature_definition_type"):
            feature_definition_type = FeatureDefinitionType.from_proto_value(
                the_proto.feature_definition_type
            )
        return cls(
            online_feature_tables=online_fts,
            feature_definition_type=feature_definition_type,
        )

    @classmethod
    def load(cls, path: str):
        """
        Loads a binary serialized ProtoFeatureTablesForServing protocol buffer.

        :param path: Root path to the binary file.
        :return: :py:class:`~databricks.ml_features_common.entities.feature_tables_for_serving.FeatureTablesForServing`
        """
        proto = ProtoFeatureTablesForServing()
        # The load path may be None when the model is packaged by Feature Store, but did not use any
        # feature tables (eg just feature functions)
        if not path:
            return cls.from_proto(proto)
        with open(os.path.join(path, cls.DATA_FILE), "rb") as f:
            proto.ParseFromString(f.read())
        return cls.from_proto(proto)



from functools import reduce
from typing import Any, Dict, List, Optional

import pandas as pd
from gevent.pool import Pool

from databricks.feature_store.entities.feature_table_metadata import (
    FeatureTableMetadata,
)
from databricks.feature_store.feature_functions.feature_function_executor import (
    FeatureFunctionExecutor,
)
from databricks.feature_store.online_lookup_client import OnlineLookupClient
from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.feature_store.utils.lakebase_utils import (
    get_lakebase_connection_pool_size,
)
from databricks.feature_store.utils.latency_debugger import LatencyDebugger
from databricks.feature_store.utils.lookup_key_utils import LookupKeyType
from databricks.ml_features_common.entities.data_type import DataType
from databricks.ml_features_common.entities.feature_column_info import FeatureColumnInfo
from databricks.ml_features_common.entities.feature_spec import FeatureSpec
from databricks.ml_features_common.entities.on_demand_column_info import (
    OnDemandColumnInfo,
)
from databricks.ml_features_common.entities.online_feature_table import (
    PrimaryKeyDetails,
)
from databricks.ml_features_common.utils.data_type_utils import (
    deserialize_default_value_primitive_type,
)
from databricks.ml_features_common.utils.feature_spec_utils import (
    COLUMN_INFO_TYPE_FEATURE,
    COLUMN_INFO_TYPE_ON_DEMAND,
    COLUMN_INFO_TYPE_SOURCE,
    get_feature_execution_groups,
)


class _PgTableFetcher:
    """
    Each (feature_table, lookup_keys) pair is associated with a single _PgTableFetcher instance.
    The fetcher is responsible for fetching data from the online store and filling in default values.
    """

    def __init__(
        self,
        lookup_key: LookupKeyType,
        primary_keys: List[PrimaryKeyDetails],
        feature_column_infos: List[FeatureColumnInfo],
        lookup_client: OnlineLookupClient,
        feature_type_mapping: Dict[str, DataType],
        latency_debugger: LatencyDebugger,
    ):
        """
        Args:
            lookup_key: A list of input column names. The ordering needs to match the ordering of the primary keys.
            primary_keys: Primary keys defined in the online table
            feature_columns: A list of feature columns to fetch
            lookup_client: Lookup client for the online table
            feature_type_mapping: Mapping from feature name to data type.
        """
        self.lookup_key = lookup_key
        self.lookup_client = lookup_client
        self.feature_column_infos = feature_column_infos
        self._primary_key_names = [pk.name for pk in primary_keys]
        self._feature_names = [fci.feature_name for fci in feature_column_infos]
        self._latency_debugger = latency_debugger
        self._default_values = {
            fci.feature_name: deserialize_default_value_primitive_type(
                fci.default_value_str, feature_type_mapping[fci.feature_name]
            )
            for fci in feature_column_infos
            if fci.default_value_str is not None
        }
        self._renaming = {
            fci.feature_name: fci.output_name
            for fci in feature_column_infos
            if fci.feature_name != fci.output_name
        }

    def _build_lookup_key_list(self, rows: List[Dict[str, Any]]) -> LookupKeyList:
        return LookupKeyList(
            columns=self._primary_key_names,
            rows=[[row[lookup_key] for lookup_key in self.lookup_key] for row in rows],
        )

    def fetch(self, rows: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        lookup_key_list = self._build_lookup_key_list(rows)
        self._latency_debugger.add_latency("pg_build_lookup_key_list")
        lookup_results = self.lookup_client.lookup_feature_dicts(
            lookup_key_list, self._feature_names
        )
        self._latency_debugger.add_latency("pg_lookup_feature_dicts")
        # Fill in default values.
        for i, lookup_result in enumerate(lookup_results):
            for feature_name, default_value in self._default_values.items():
                if lookup_result.get(feature_name, None) is None:
                    lookup_results[i][feature_name] = default_value
        # Rename columns.
        if self._renaming:
            for i, lookup_result in enumerate(lookup_results):
                lookup_results[i] = {
                    self._renaming.get(k, k): v for k, v in lookup_result.items()
                }
        return lookup_results


class PostgresFetcher:
    def __init__(
        self,
        feature_spec: FeatureSpec,
        lookup_clients: Dict[str, OnlineLookupClient],
        ft_metadatas: Dict[str, FeatureTableMetadata],
        feature_function_executor: Optional[FeatureFunctionExecutor],
        latency_debugger: LatencyDebugger,
    ):
        """
        Initialize the PostgresFetcher.
        Args:
            feature_spec: FeatureSpec
            lookup_clients: feature table name -> lookup client
            ft_metadatas: feature table name -> feature_table_metadata
        """
        self.feature_spec = feature_spec
        self.feature_function_executor = feature_function_executor
        # FeatureTable (name, lookup_key) -> fetcher
        # lookup_key is a list of input column names.
        self.table_fetchers = {}
        # The desired output columns in the final result. Other columns should be filtered out.
        self._output_names = [
            ci.output_name for ci in feature_spec.column_infos if ci.include
        ]
        # All columns in the feature spec. Including input and output columns.
        self._all_column_names = [ci.output_name for ci in feature_spec.column_infos]

        self._needs_dag_execution = (
            feature_spec.has_dag() or len(feature_spec.on_demand_column_infos) > 0
        )
        # Execution group is calcualted based on input columns and the feature spec.
        # cached execution groups are used if the input columns are the same.
        self._execution_groups = None
        self._input_columns = None
        self._latency_debugger = latency_debugger
        self._lakebase_connection_pool_size = get_lakebase_connection_pool_size(
            len(ft_metadatas)
        )
        for ft, ft_metadata in ft_metadatas.items():
            for (
                lookup_key,
                feature_column_infos,
            ) in ft_metadata.feature_col_infos_by_lookup_key.items():
                self.table_fetchers[(ft, lookup_key)] = _PgTableFetcher(
                    lookup_key,
                    ft_metadata.online_ft.primary_keys,
                    feature_column_infos,
                    lookup_clients[ft],
                    ft_metadata.get_table_feature_type_mapping(),
                    latency_debugger,
                )

    def _lookup_feature_group(
        self, input_rows: List[Dict[str, Any]], features: List[FeatureColumnInfo]
    ) -> List[dict[str, Any]]:
        """
        Lookup features in an execution group.
        """
        # 1. group features by (feature_table, lookup_key)
        feature_groups = {}
        for feature in features:
            # Convert lookup_key list to tuple for use as dictionary key
            lookup_key_tuple = tuple(feature.lookup_key)
            group_key = (feature.table_name, lookup_key_tuple)

            if group_key not in feature_groups:
                feature_groups[group_key] = []
            feature_groups[group_key].append(feature)

        # 2. lookup features in each group
        group_results = []
        for (table_name, lookup_key_tuple), group_features in feature_groups.items():
            # Get the appropriate table fetcher
            fetcher = self.table_fetchers.get((table_name, lookup_key_tuple))
            if fetcher is None:
                raise ValueError(
                    f"No fetcher found for table {table_name} with lookup key {lookup_key_tuple}"
                )

            # Fetch features for this group
            group_result = fetcher.fetch(input_rows)
            group_results.append(group_result)

        # 3. merge results by rows
        if not group_results:
            # If no feature groups, return input rows as-is
            return input_rows

        # Merge all group results with input rows
        merged_results = []
        for i in range(len(input_rows)):
            merged_row = {}
            # merge features from all groups for this row
            for group_result in group_results:
                if i < len(group_result):
                    merged_row.update(group_result[i])
            # override with values from input_rows (input values take precedence)
            merged_row.update(input_rows[i])
            merged_results.append(merged_row)

        # 4. return the result
        return merged_results

    def _lookup_with_dag(self, df: pd.DataFrame) -> List[dict[str, Any]]:
        """
        Lookup features with a DAG.
        """
        if self._execution_groups is None or self._input_columns != df.columns.tolist():
            self._input_columns = df.columns.tolist()
            self._execution_groups = get_feature_execution_groups(self.feature_spec, df)

        # Start with input data
        current_rows = df.to_dict(orient="records")

        for group in self._execution_groups:
            if group.type == COLUMN_INFO_TYPE_SOURCE:
                # no action required for source columns
                continue
            elif group.type == COLUMN_INFO_TYPE_FEATURE:
                # Lookup features for this execution group
                current_rows = self._lookup_feature_group(current_rows, group.features)
            elif group.type == COLUMN_INFO_TYPE_ON_DEMAND:
                # Apply functions for this execution group
                current_rows = (
                    self.feature_function_executor.execute_functions_for_rows(
                        current_rows, group.features
                    )
                )

        return current_rows

    def lookup_features(self, df: pd.DataFrame) -> List[dict[str, Any]]:
        """
        Lookup features.

        Args:
            df: DataFrame with lookup keys.

        Returns:
            List of dictionaries with features.
        """
        if self._needs_dag_execution:
            all_results = self._lookup_with_dag(df)
        else:
            input_rows = df.to_dict(orient="records")
            fetchers = list(self.table_fetchers.values())

            # Use greenlet thread pool only if there are multiple fetchers
            if len(fetchers) > 1:
                # Use greenlet pool for parallel execution. The pool size should be equal
                # or greater than the size of the connection pool for fully use the
                # connections.
                pool = Pool(size=self._lakebase_connection_pool_size)
                results = pool.map(lambda fetcher: fetcher.fetch(input_rows), fetchers)
                pool.join()
            else:
                # Single fetcher - execute directly without thread pool
                results = [fetchers[0].fetch(input_rows)] if fetchers else []

            # Merge all results by rows.
            all_results = [
                reduce(lambda a, b: {**a, **b}, group, {})
                for group in zip(*results, input_rows)
            ]
        if all_results and len(self._all_column_names) == len(self._output_names):
            return all_results
        else:
            # TODO[ML-55717]: Optimize the filtering performance.
            # filter out the columns that are excluded.
            return [
                {k: v for k, v in result.items() if k in self._output_names}
                for result in all_results
            ]

    def get_model_input_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Get the model input DataFrame.
        """
        feature_values = self.lookup_features(df)
        # convert the feature_values to a DataFrame
        return pd.DataFrame(
            feature_values,
            columns=self._output_names,  # specify the column ordering
        )



import os
from importlib.machinery import SourceFileLoader

from setuptools import find_namespace_packages, setup

from databricks.ml_features_common import mlflow_model_constants

version = (
    SourceFileLoader(
        "databricks.feature_store.feature_lookup_version",
        os.path.join("databricks", "feature_store", "feature_lookup_version.py"),
    )
    .load_module()
    .VERSION
)

with open("README.md") as f:
    readme_contents = f.read()

with open("LICENSE.md") as f:
    license_contents = f.read()

# Append the text from LICENSE.md to the end of the README.md content
readme_contents_with_license = (
    readme_contents + "\n\n" + "# License" + "\n\n" + license_contents
)

# Overwrite the existing README.md file with the readme + license content for the setup() step.
# The original version will be restored to it's original state after setup() finishes.
with open("README.md", "w") as f:
    f.write(readme_contents_with_license)


# TODO (ML-20365): DRY this with core client into universe/feature-store-commons
def remove_comments_and_empty_lines(lines):
    def is_comment_or_empty(line):
        stripped = line.strip()
        return stripped == "" or stripped.startswith("#")

    return [line for line in lines if not is_comment_or_empty(line)]


with open("requirements.txt", "r") as f:
    REQUIREMENTS = remove_comments_and_empty_lines(f.readlines())

setup(
    name=mlflow_model_constants.FEATURE_LOOKUP_CLIENT_PIP_PACKAGE,
    version=version,
    author_email="feedback@databricks.com",
    license="Databricks Proprietary License",
    license_files=("LICENSE.md", "NOTICE.md"),
    long_description=readme_contents_with_license,
    long_description_content_type="text/markdown",
    # The minimum required python version should align with the conda env specified by feature store client log_model()
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    packages=find_namespace_packages(include=["databricks.*"], exclude=["tests"]),
    include_package_data=True,
    package_data={
        "databricks.feature_store.utils": ["headers.yaml"],
    },
    install_requires=REQUIREMENTS,
    author="Databricks",
    description="Databricks Feature Store Feature Lookup Client",
)

# Restore the README.md file to it's original state
with open("README.md", "w") as f:
    f.write(readme_contents)



import json
from typing import Any, List

from azure.cosmos.exceptions import CosmosResourceNotFoundError

PARTITION_KEY = "partitionKey"
PATHS = "paths"
PRIMARY_KEY_PROPERTY_NAME_VALUE = "_feature_store_internal__primary_keys"
FEATURE_STORE_SENTINEL_ID_VALUE = "_fs"

DEFAULT_RETRY_CONFIG = {"retry_total": 3, "timeout": 30}


def to_cosmosdb_primary_key(primary_key_values: List[Any]):
    return json.dumps(primary_key_values)


def is_not_found_error(e: Exception) -> bool:
    # Cosmos DB clients are lazily constructed without validation. This results in ContainerProxy.read_item potentially
    # throwing if the account, database, container, or item doesn't exist.
    # Determine if the CosmosResourceNotFoundError was caused by ContainerProxy.read_item finding no result.
    return isinstance(e, CosmosResourceNotFoundError) and e.status_code == 404



"""
Test utilities shared between the lookup-client and the core client
"""

# TODO (ML-30526): Move this file to the tests module.


def mock_context_manager(mock_context_mgr, resource):
    """
    Modify the mock context manager to return the given resource when it's __enter__() protocol method is invoked.
    """
    mock_context_mgr.__enter__.return_value = resource



from databricks.feature_store.lookup_engine.lookup_cosmosdb_engine import (
    LookupCosmosDbEngine,
)
from databricks.feature_store.lookup_engine.lookup_dynamodb_engine import (
    AwsAccessKey,
    LookupDynamoDbEngine,
)
from databricks.feature_store.lookup_engine.lookup_engine import LookupEngine
from databricks.feature_store.lookup_engine.lookup_lakebase_engine import (
    LookupLakebaseEngine,
)
from databricks.feature_store.lookup_engine.lookup_legacy_brickstore_http_engine import (
    LookupLegacyBrickstoreHttpEngine,
)
from databricks.feature_store.lookup_engine.lookup_mysql_engine import LookupMySqlEngine
from databricks.feature_store.lookup_engine.lookup_sql_engine import LookupSqlEngine
from databricks.feature_store.lookup_engine.lookup_sql_server_engine import (
    LookupSqlServerEngine,
)

__all__ = [
    "LookupEngine",
    "LookupSqlEngine",
    "LookupMySqlEngine",
    "LookupSqlServerEngine",
    "LookupDynamoDbEngine",
    "LookupCosmosDbEngine",
    "LookupLakebaseEngine",
    "LookupLegacyBrickstoreHttpEngine",
]



import logging
import os
import random
import time
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Tuple, Union

import jwt
import numpy as np
import psycopg
import sqlalchemy
from sqlalchemy import bindparam, column, select, table, tuple_
from sqlalchemy.sql.elements import quoted_name

from databricks.feature_store.lookup_engine.async_refill_engine import AsyncRefillEngine
from databricks.feature_store.lookup_engine.lookup_sql_engine import LookupSqlEngine
from databricks.feature_store.lookup_engine.oauth_token_manager import OAuthTokenManager
from databricks.feature_store.metrics.feature_store_metrics_recorder import (
    METRIC_POSTGRES_CONNECTION_ACQUISITION_LATENCY,
    METRIC_POSTGRES_QUERY_LATENCY,
    get_metrics_recorder,
)
from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.feature_store.utils.lakebase_constants import (
    LAKEBASE_OAUTH_TOKEN_FILE_PATH,
)
from databricks.feature_store.utils.logging_utils import get_logger
from databricks.feature_store.utils.metrics_utils import LookupClientMetrics
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)

# TODO[ML-60213]: move these env vars to lookup_client_envvars.py
ENV_LAKEBASE_CONNECTION_RECYCLE_SECONDS = (
    "FEATURE_SERVING_LAKEBASE_CONNECTION_RECYCLE_SECONDS"
)
ENV_CONNECTION_WARMING_INTERVAL_SECONDS = (
    "FEATURE_SERVING_CONNECTION_WARMING_INTERVAL_SECONDS"
)
ENV_QUERY_LATENCY_WARNING_THRESHOLD_MS = (
    "FEATURE_SERVING_QUERY_LATENCY_WARNING_THRESHOLD_MS"
)

FIELD_NAME_FOR_ROLE_IN_TOKEN = "sub"
JWT_ALGORITHM = "RS256"
BASE_URL = "postgresql+psycopg://"
QUERY_MAX_ATTEMPTS = 2

_logger = get_logger(__name__, log_level=logging.INFO)


class LookupLakebaseEngine(LookupSqlEngine):
    # class-level SQLAlchemy engine cache shared by all LookupLakebaseEngine instances in the same process
    _engine_cache = {}

    def __init__(
        self,
        online_feature_table: OnlineFeatureTable,
        ro_user: str,
        ro_password: str,
        connection_pool_size: int,
    ):
        # The oauth token is refreshed every 30mins and expires 1 hour. Recycle the connection
        # every 15mins to pickup the latest token.
        self.connection_recycle_seconds = int(
            os.environ.get(ENV_LAKEBASE_CONNECTION_RECYCLE_SECONDS, "900")
        )
        # If the SQL query latency is greater than this threshold, log a warning.
        self.query_latency_warning_threshold_ms = int(
            os.environ.get(ENV_QUERY_LATENCY_WARNING_THRESHOLD_MS, "300")
        )
        # Keep the connection pool warm by periodically recycling connections. This is
        # needed when the endpoint is idle and no real requests to trigger the recycling.
        self.connection_warming_interval_seconds = int(
            os.environ.get(ENV_CONNECTION_WARMING_INTERVAL_SECONDS, "5")
        )

        # The size of the connection pool.
        self.connection_pool_size = connection_pool_size
        # Initialize query cache dictionary
        self.queries = {}
        self._oauth_token_manager = OAuthTokenManager(
            oauth_token_file_path=LAKEBASE_OAUTH_TOKEN_FILE_PATH,
            password_override=ro_password,
        )
        # The parent constructor calls get_connection which requires the oauth token to be set.
        # So we need to set the oauth token manager before calling super().__init__
        super().__init__(online_feature_table, ro_user, ro_password)
        self._oauth_token_manager.start_token_refresh_thread()
        self._initialize_engine()

    @staticmethod
    def _current_time_ms():
        """
        Returns the current time in milliseconds.
        This method can be patched in tests.
        """
        return int(time.time() * 1000)

    def _get_engine_cache_key(self):
        return (self.host, self.port, self.database_name)

    def _initialize_engine(self):
        engine_cache_key = self._get_engine_cache_key()
        if engine_cache_key not in self._engine_cache:
            self._engine_cache[engine_cache_key] = self._create_new_engine()

    def _create_new_engine(self):
        """Create a new AsyncRefillEngine instance."""
        pool_recycle = self.connection_recycle_seconds
        # Add randomization to distribute connection recycling
        pool_recycle_with_jitter = pool_recycle + random.randint(0, 10)

        new_engine = AsyncRefillEngine(
            pool_size=self.connection_pool_size,
            pool_recycle=pool_recycle_with_jitter,
            pool_warming_interval=self.connection_warming_interval_seconds,
            creator=self._connect,
        )
        return new_engine

    # Override
    def _get_database_and_table_name(
        self, online_table_name
    ) -> Tuple[str, Optional[str], str]:
        name_components = online_table_name.split(".")
        if len(name_components) != 3:
            raise ValueError(
                f"Online table name {online_table_name} is misformatted and must be in 3L format for Lakebase stores"
            )
        return (name_components[0], name_components[1], name_components[2])

    # Override
    def is_lakebase_engine(self) -> bool:
        return True

    # Lakebase sql connection uses a connection pool
    # Override
    @contextmanager
    def _get_connection(self):
        engine = self._engine_cache[self._get_engine_cache_key()]
        request_connection_time = self._current_time_ms()
        connection = engine.acquire()
        get_connection_time_ms = self._current_time_ms() - request_connection_time
        get_metrics_recorder().record_histogram(
            METRIC_POSTGRES_CONNECTION_ACQUISITION_LATENCY,
            get_connection_time_ms,
        )
        if get_connection_time_ms > self.query_latency_warning_threshold_ms:
            _logger.warning(f"Slow connection acquisition: {get_connection_time_ms}ms;")
        try:
            # When the caller invokes "with _get_connection() as x", the connection will be returned as "x"
            yield connection
        except Exception as e:
            connection.invalidate()
            raise e
        finally:
            # Everything below here will be executed in contextmanager.__exit__()
            # Release the connection and return it back to the pool
            engine.release(connection)

    def _connect(self):
        oauth_token_or_password = (
            self._oauth_token_manager.get_oauth_token_or_password()
        )
        # self.user is parsed from EnvVar. If not set, parse the client_id
        # from the oauth token
        db_user = self.user
        if not db_user:
            content = jwt.decode(
                oauth_token_or_password,
                algorithms=[JWT_ALGORITHM],
                # No worry, the token is validated by Postgres
                options={"verify_signature": False},
            )
            db_user = content[FIELD_NAME_FOR_ROLE_IN_TOKEN]

        return psycopg.connect(
            host=self.host,
            port=self.port,
            dbname=self.database_name,
            user=db_user,
            password=oauth_token_or_password,
            sslmode="require",
        )

    # Override
    def _database_contains_feature_table(self, sql_connection):
        # TODO[ML-53997]: implement validation
        return True

    # Override
    def _database_contains_primary_keys(self, sql_connection):
        # TODO[ML-53997]: implement validation
        return True

    # Override
    @classmethod
    def _sql_safe_name(cls, name):
        return name

    # Override with batch optimization for lakebase engine
    def _get_sql_results(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
        metrics: LookupClientMetrics = None,  # Ignored for lakebase engine
    ) -> List[Union[Dict[str, Any], np.ndarray, sqlalchemy.engine.Row]]:
        # For empty lookup list, return empty results
        if not lookup_list.rows:
            return []

        # Create cache key from columns and feature names (including whether single or multi-key)
        is_single_key = len(lookup_list.columns) == 1
        cache_key = (tuple(lookup_list.columns), tuple(feature_names), is_single_key)

        # Check if we have a cached query for this combination
        if cache_key in self.queries:
            query = self.queries[cache_key]
        else:
            # Create table reference with schema
            quoted_table_name = quoted_name(self.table_name, quote=True)
            quoted_schema_name = (
                quoted_name(self.schema_name, quote=True) if self.schema_name else None
            )
            table_ref = table(quoted_table_name, schema=quoted_schema_name)

            # Create column references
            pk_columns = [
                column(quoted_name(pk, quote=True)) for pk in lookup_list.columns
            ]
            feature_columns = [
                column(quoted_name(f, quote=True)) for f in feature_names
            ]

            # Build the SELECT query with WHERE clause template
            base_query = select(*pk_columns, *feature_columns).select_from(table_ref)

            # Add WHERE clause with parameter binding template
            if is_single_key:
                # Single primary key: WHERE pk IN (:values)
                pk_col = pk_columns[0]
                query = base_query.where(
                    pk_col.in_(bindparam("values", expanding=True))
                )
            else:
                # Multiple primary keys: WHERE (pk1, pk2) IN (:value_tuples)
                pk_tuple = tuple_(*pk_columns)
                query = base_query.where(
                    pk_tuple.in_(bindparam("value_tuples", expanding=True))
                )

            # Cache the complete query for future use
            self.queries[cache_key] = query

        # Prepare parameters for the cached query
        if is_single_key:
            values = [row[0] for row in lookup_list.rows]
            params = {"values": values}
        else:
            value_tuples = [tuple(row) for row in lookup_list.rows]
            params = {"value_tuples": value_tuples}

        def execute_query():
            with self._get_connection() as sql_connection:
                query_start_time = self._current_time_ms()
                # TODO[ML-58491]: explore preparing the statement and commit. Then only use a
                #    statement id for better performance.
                sql_data = sql_connection.execute(query, params)
                # Rows in results may not be in the same order as the lookup list.
                rows = sql_data.fetchall()
                query_latency = self._current_time_ms() - query_start_time
                get_metrics_recorder().record_histogram(
                    METRIC_POSTGRES_QUERY_LATENCY,
                    query_latency,
                )
                return rows

        # Retry on exceptions up to QUERY_MAX_ATTEMPTS
        for attempt in range(QUERY_MAX_ATTEMPTS):
            try:
                results = execute_query()
                break
            except Exception as e:
                if attempt < QUERY_MAX_ATTEMPTS - 1:
                    _logger.warning(f"Retrying query for reason: {str(e)}")
                else:
                    # The last attempt failed too.
                    _logger.error(f"Last execute_query attempt failed")
                    raise e
        # Create a mapping from primary key values to result rows for fast lookup
        pk_to_result = {}
        for result_row in results:
            # Extract primary key values from the result (first len(lookup_list.columns) columns)
            pk_values = tuple(result_row[: len(lookup_list.columns)])
            # Extract feature values (remaining columns)
            feature_values = result_row[len(lookup_list.columns) :]
            pk_to_result[pk_values] = feature_values

        # Build ordered results matching the original lookup order
        # TODO[ML-58492]: try not to reorder the results for FeatureServing because it's not
        #    necessary for JSON responses. It would matter for model serving.
        ordered_results = []
        for row in lookup_list.rows:
            pk_tuple = tuple(row)
            if pk_tuple in pk_to_result:
                # Create a dictionary directly
                feature_values = pk_to_result[pk_tuple]
                ordered_results.append(dict(zip(feature_names, feature_values)))
            else:
                # No result found for this primary key - return None values
                _logger.warning(
                    f"No feature values found in {self.table_name} for primary key {dict(zip(lookup_list.columns, row))}."
                )
                ordered_results.append({name: None for name in feature_names})

        return ordered_results



import logging
import time

import gevent
from gevent.lock import RLock, Semaphore
from gevent.queue import Empty, Queue
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool

from databricks.feature_store.utils.logging_utils import get_logger

_logger = get_logger(__name__, log_level=logging.INFO)
BASE_URL = "postgresql+psycopg://"


class AsyncRefillEngine:
    """
    A PostgreSQL engine that maintains pool of connections and actively refills
    the pool when connections expire.

    Unlike the standard SQLAlchemy engine, the connection pool in this engine
    usually does not create new connections in the critical query path. Instead,
    when aquiring a connection,it skip and discard expired connections from the
    head of the queue to get the next available connection as soon as possible.
    Then it refills the pool with new connections asynchronously.

    The engine also starts a background greenlet that keeps acquiring and
    releasing connections to keep the pool warm even when there is no active
    query.

    Gevent compatible: this engine works in a gevent environment. If not, the
    async operations won't start.

    Usage:
        Use `acquire` to get a connection from the pool and `release` to return
        it back to the pool. Calling close on the connection object will not
        return it back to the pool.
    """

    def __init__(
        self,
        pool_size,
        pool_recycle,
        pool_warming_interval=5,
        pool_timeout=2,
        pool_init_gap=2,
        creator=None,
    ):
        """
        Initialize an AsyncRefillEngine.

        :param pool_size: The size of the pool.
        :param pool_recycle: The time in seconds before the connection expires.
        :param pool_warming_interval: The interval in seconds at which the pool
          warming thread runs. In the pool warming thread, we recycle connections
          3x of pool_warming_interval before the expiration. So this number
          should be much smaller than 1/3 of pool_recycle to keep the connections
          alive longer. But an extremely small interval can also cause too many
          requests to the connection pool.
        :param pool_timeout: The timeout in seconds for getting a connection from
          the pool.
        :param pool_init_gap: The gap in seconds between initializing connections.
          This is needed to avoid connections expiring at the same time. Also to
          avoid too many requests to hit the auth rate limit for creating
          connections during initialization.
        :param creator: The creator function to use to create the connection.
        """
        self._engine = create_engine(BASE_URL, creator=creator, poolclass=NullPool)
        self._pool = Queue(maxsize=pool_size)
        # Lock for the filed _in_use. Any access to _in_use must be protected by this
        self._in_use_count_lock = Semaphore()
        # Lock for the field _refill_count. Any access to _refill_count must be
        self._refill_count_lock = Semaphore()
        # Reentrant lock for the acquire method. acquire needs to be thread safe
        # because it modifies some internal state of this class. Using RLock because
        # acquire can be called recursively.
        self._connection_acquire_lock = RLock()
        self._pool_size = pool_size
        self._pool_recycle = pool_recycle
        self._pool_timeout = pool_timeout
        self._pool_warming_interval = pool_warming_interval
        # How long before the connection expires should the pool warming thread
        # recycle the connection. this makes sure there is at least one connection
        # in the pool can last until the next warming cycle.
        #
        # Choosing 3x because in production there are at most 2 connections checked
        # out at the same time. Keeping one extra connection for edge cases. This might
        # needs tuning when we support parallel queries for multiple tables.
        self._pre_recycle_time = pool_warming_interval * 3
        # Number of connections currently checked out
        self._in_use = 0
        # Number of connections currently being refilled
        self._refill_count = 0

        # Warm-up pool
        while self._pool.qsize() < self._pool_size:
            try:
                conn = None
                conn = self._create_conn()
                self._pool.put(conn)
            except Exception as e:
                _logger.warning(f"[pool] Retrying to init connection because: {e}")
                if conn:
                    try:
                        conn.close()
                    except Exception:
                        pass
            # Initialize the connections with gaps to avoid expiring them all
            # together.
            gevent.sleep(pool_init_gap)
        if pool_warming_interval > 0:
            gevent.spawn(self._pool_warming_thread)

    def _create_conn(self):
        conn = self._engine.connect()
        conn.starttime = time.time()
        return conn

    def _pool_warming_thread(self):
        """
        The pool warming thread is to keep the pool warm when the endpoint is
        not being queried.
        """
        while True:
            try:
                conn = self._acquire_connection_internal(is_warming=True)
                if conn:
                    self.release(conn)
            except Exception as e:
                _logger.warning(f"[pool] Error warming up connection: {e}")
            gevent.sleep(self._pool_warming_interval)

    def _is_valid(self, conn, is_warming=False):
        """
        Check if the connection is valid.
        The conn.invalidated field and the starttime both are indicators for
        connection to be recycled. But they are handled differently:
        - When conn.invalidated==True, it means the connection is broken and
            cannot be used at all and we must recycle it immediately.
        - But for starttime, we have 2 expiration time for it. The first one
            is a soft expiration time hinting the pool warming thread to
            recycle it before it actually expires. The second is a hard
            deadline meaning it has to be recycled by either a real request
            or a pool warming thread. Based on is_warming flag, we can determine
            which expiration time to use.
        """
        # First check if the connection is invalidated. If it is, return False.
        if conn.invalidated:
            return False
        # Validate the connection expiration based on if the check is for a
        # connection warming request. For connection warming request, we
        # recycle the connections sooner than normal to make sure a real
        # request doesn't stuck on a stale connection.
        start = getattr(conn, "starttime", None)
        if start and self._pool_recycle > 0:
            recycle_time = (
                self._pool_recycle - self._pre_recycle_time
                if is_warming
                else self._pool_recycle
            )
            if time.time() - start > recycle_time:
                return False
        return True

    def acquire(self):
        """
        Acquires a connection from the pool. This method is greenlet thread-safe. It
        blocks until a connection is available.

        :return: A connection from the pool.
        """
        return self._acquire_connection_internal(is_warming=False)

    def _acquire_connection_internal(self, is_warming=False):
        """
        :param is_warming: Whether the connection is being acquired by the pool warming thread.
            If True, it will not guarantee to return a connection. Instead it may return None
            if the connection queue is almost empty.
        """
        with self._connection_acquire_lock:
            if (
                self._pool.qsize() + self.checkedout() + self.refilling()
                < self._pool_size
            ):
                # There are not enough connections in the managed connection pool. This is not
                # expected to happen in normal case but it could happen if any creation of
                # connections during refilling failed. Spawn a refill task immediately to create
                # a new connection in the background.
                gevent.spawn(self._refill)

            # Now we know there is at least one connection guaranteed to be available soon.
            # beacuse we spawn a refill task above when there are not enough connections.
            # Try to get a connection from the queue. If it's invalid, dispose it and try
            # again recursively.
            try:
                conn = self._pool.get(timeout=self._pool_timeout)
            except Empty:
                _logger.error(
                    f"[pool] timeout getting connection. checked-in:{self.checkedin()}, checked-out:{self.checkedout()}, refilling:{self.refilling()}"
                )
                raise TimeoutError(f"No available connection in {self._pool_timeout}s")

            if not self._is_valid(conn, is_warming):
                # Invalid connections need to be disposed.

                # First trigger a refill to start creating the replacement connection
                # in the background.
                gevent.spawn(self._refill)

                if is_warming and self.checkedin() == 0:
                    # If in the pool warming thread and this is the only connection in the
                    # queue, put it back immediately to avoid blocking the real requests.
                    # The connection might still be valid because _is_valid for warming
                    # thread has a shorter expiration time.
                    # This will lead to one extra connection in total than the pool size.
                    # (because we spawn a new _refill but not closing this connection) but
                    # it's okay because the _put_in_queue will discard any extra
                    # connections.
                    self._put_in_queue(conn)
                    return None
                else:
                    try:
                        conn.close()
                    except Exception:
                        pass
                # Try to acquire a connection again.
                return self._acquire_connection_internal(is_warming=is_warming)

            with self._in_use_count_lock:
                self._in_use += 1
            return conn

    def _put_in_queue(self, conn):
        try:
            self._pool.put_nowait(conn)
        except Exception:
            try:
                conn.close()
            except Exception:
                pass

    def release(self, conn):
        with self._in_use_count_lock:
            self._in_use -= 1
        self._put_in_queue(conn)

    def _refill(self):
        try:
            with self._refill_count_lock:
                self._refill_count += 1
            conn = self._create_conn()
            self._put_in_queue(conn)
        except Exception as e:
            _logger.error(f"[pool] async refill failed: {e}")
        finally:
            with self._refill_count_lock:
                self._refill_count -= 1

    # Observability
    # How many connections are in the pool.
    def checkedin(self):
        return self._pool.qsize()

    # How many connections are checked out and being used.
    def checkedout(self):
        with self._in_use_count_lock:
            return self._in_use

    # How many connections are currently being refilled.
    def refilling(self):
        with self._refill_count_lock:
            return self._refill_count



from databricks.ml_features_common.entities._feature_store_object import (
    _FeatureStoreObject,
)
from databricks.ml_features_common.entities.feature_spec_constants import (
    FEATURE_STORE_CLIENT_VERSION_FIELD_NAME,
    SERIALIZATION_VERSION,
    WORKSPACE_ID_FIELD_NAME,
)
from databricks.ml_features_common.utils.yaml_utils import write_yaml


class _FeatureSpecBase(_FeatureStoreObject):

    FEATURE_ARTIFACT_FILE = "feature_spec.yaml"
    SERIALIZATION_VERSION_NUMBER = 11

    def __init__(
        self,
        serialization_version: int,
        workspace_id: int,
        feature_store_client_version: str,
    ):
        self._serialization_version = serialization_version
        self._workspace_id = workspace_id
        self._feature_store_client_version = feature_store_client_version

    @property
    def serialization_version(self) -> int:
        return self._serialization_version

    @property
    def workspace_id(self) -> int:
        return self._workspace_id

    @property
    def feature_store_client_version(self) -> str:
        return self._feature_store_client_version

    def _to_dict(self):
        raise NotImplementedError("Derived classes must implement this method")

    def save(self, path: str):
        """
        Convert spec to a YAML artifact and store at given `path` location.
        :param path: Root path to where YAML artifact is expected to be stored.
        :return: None
        """
        write_yaml(
            path=path,
            file_name=self.FEATURE_ARTIFACT_FILE,
            data_dict=self._to_dict(),
        )



import re

from databricks.ml_features_common.entities.data_type import DataType

ELEMENT_TYPE = "elementType"
KEY_TYPE = "keyType"
VALUE_TYPE = "valueType"


def get_data_type_from_details(details):
    if isinstance(details, str):
        return DataType.from_string(details)
    return DataType.from_string(details.get("type"))


def parse_decimal_details(details):
    """
    Parse Spark's DecimalType JSON representation for precision and scale.

    :param details: DecimalType JSON representation e.g. "decimal(5,3)" has precision 5 and scale 3.
    """
    match = re.search("^decimal\((\d+),(\d+)\)$", details)
    if match is None:
        raise Exception(f"Malformed decimal data type details {details}")

    return int(match.group(1)), int(match.group(2))



from typing import Any, Callable, Dict, List, Set

import pandas as pd

from databricks.feature_store.entities.feature_functions_for_serving import (
    FeatureFunctionForServing,
    FeatureFunctionsForServing,
)
from databricks.feature_store.feature_functions.feature_function_loader import (
    FeatureFunctionLoader,
)
from databricks.feature_store.utils import feature_function_type_utils
from databricks.ml_features_common.entities.on_demand_column_info import (
    OnDemandColumnInfo,
)

OutputNameType = str


class _MaterializableFeatureFunction:
    """
    Encapsulates data required to evaluate a Feature Function, obtained from FeatureSpec and
    FeatureFunctionsForServing.
    """

    def __init__(
        self,
        ff_for_serving: FeatureFunctionForServing,
        odci: OnDemandColumnInfo,
        fn: Callable,
    ):
        self.odci = odci
        # fn is a Python function which can be called on UDF inputs to produce the uncasted UDF
        # output.
        self.fn = fn
        # column_args represents the list of columns that should be passed to fn.
        # For example, given a row of materialized feature columns, we evaluate the feature function by calling:
        #   fn(*[row[arg] for arg in column_args])
        self.column_args = [
            odci.input_bindings[input_param.name]
            for input_param in ff_for_serving.input_params
        ]
        self.return_data_type = ff_for_serving.data_type


class FeatureFunctionExecutor:
    """
    Library to execute Feature Functions on a pandas DataFrame in Databricks Model Serving.
    """

    def __init__(
        self,
        feature_functions_for_serving: FeatureFunctionsForServing,
        on_demand_column_infos: List[OnDemandColumnInfo],
    ):
        self.materializable_features = self._get_materializable_feature_metadata(
            feature_functions_for_serving, on_demand_column_infos
        )

    def execute_functions_for_rows(
        self, rows: List[Dict[str, Any]], functions_to_apply: List[OnDemandColumnInfo]
    ) -> List[Dict[str, Any]]:
        """
        Evaluates Feature Functions on a list of rows.
        """
        output_feat = [
            (info.output_name, self.materializable_features[info.output_name])
            for info in functions_to_apply
        ]
        for row in rows:
            for output, feat in output_feat:
                row[output] = feat.fn(*[row[arg] for arg in feat.column_args])
        return rows

    def execute_feature_functions(
        self, df: pd.DataFrame, functions_to_apply: List[OnDemandColumnInfo]
    ):
        """
        Evaluates Feature Functions on df.

        This method evaluates all Feature Functions passed to the class constructor. In the future,
        it may be extended to support calling only a subset of Feature Functions. This will be
        necessary, for example, if we support operations like TRANFORM-LOOKUP, where not all
        Feature Functions are applied in tandem.

        df is expected to include all function inputs.

        Beware that pandas may apply typecasting to df columns prior to executing the Feature
        Functions. Should the DataFrame contain columns of different type, the columns will be
        casted to the lowest common ancestor in the type tree. This is because a pd.Series is
        passed to _execute_on_demand_computation_on_row, and pd.Series is homogenously typed.

        For example, a DataFrame containing columns of np.int64 and np.float64 type will be casted
        to np.float64. A DataFrame containing columns of np.int64 and str type will be casted to
        object.

        In some cases, this may result in loss-of-precision (particularly at the large end of the
        int64 range). Should this be problematic to customers, we could alleviate this problem
        using one of the following approaches:
          a. Not use Pandas to apply Feature Functions, instead using eg dictionaries.
          b. Currying function application column-by-column, using single-column Pandas DataFrames
            or rows.

        The performance hit on these approaches is not yet clear, and probably requires profiling.

        :return: pd.DataFrame containing on-demand features. Number of columns == len(self.odci)
        """
        output_feat = [
            (info.output_name, self.materializable_features[info.output_name])
            for info in functions_to_apply
        ]

        def _execute_on_demand_computation_on_row(
            row: pd.core.series.Series,
        ) -> List[Any]:
            """
            Evaluates Feature Functions on the provided Pandas Series.
            """
            return [
                feat.fn(*[row[arg] for arg in feat.column_args])
                for output, feat in output_feat
            ]

        # Apply functions for each row of the df.
        uncasted_df = df.apply(
            _execute_on_demand_computation_on_row, axis=1, result_type="expand"
        )
        # Only select the function output columns.
        uncasted_df.columns = [info.output_name for info in functions_to_apply]
        return self._convert_output_types(uncasted_df)

    def _get_materializable_feature_metadata(
        self,
        feature_functions_for_serving: FeatureFunctionsForServing,
        on_demand_column_infos: List[OnDemandColumnInfo],
    ) -> Dict[OutputNameType, _MaterializableFeatureFunction]:
        """
        Re-structures the provided constructor inputs into a format more efficient for Feature
        Function evaluation:
          {on demand feature output name -> _MaterializableFeatureFunction}

        Every OnDemandColumnInfo UDF must be described by a FeatureFunctionForServing.

        For example:
            _get_materializable_feature_metadata(
              feature_functions_for_serving = [
                FeatureFunctionForServing(full_name="cat.sch.fn1", ...),
                FeatureFunctionForServing(full_name="cat.sch.fn2", ...),
              ],
              on_demand_column_infos = [
                OnDemandColumnInfo(udf_name="cat.sch.fn1", ...),
                OnDemandColumnInfo(udf_name="cat.sch.fn2", ...),
              ]
            )

        =>

            {
              "cat.sch.fn1":
                _MaterializableFeatureFunction(
                  FeatureFunctionForServing(full_name="cat.sch.fn1", ...),
                  OnDemandColumnInfo(udf_name="cat.sch.fn1", ...),
                  lambda x, y: ...
                ),
              "cat.sch.fn2":
                _MaterializableFeatureFunction(
                  FeatureFunctionForServing(full_name="cat.sch.fn2", ...),
                  OnDemandColumnInfo(udf_name="cat.sch.fn2", ...)
                  lambda a: ...
                )
            }
        """
        ff_for_serving_by_udf_name = {
            ff_for_serving.full_name: ff_for_serving
            for ff_for_serving in feature_functions_for_serving.ff_for_serving
        }
        fns_by_udf_name = FeatureFunctionLoader.load_feature_functions_map(
            feature_functions_for_serving
        )

        # Every OnDemandColumnInfo UDF must be described by a FeatureFunctionForServing. Because
        # multiple OnDemandColumnInfos may execute the same UDF, we have:
        # len(ff_for_serving_by_udf_name) >= len(odci_by_udf_name)
        # Return a _MaterializableFeatureFunction for each OnDemandColumnInfo
        return {
            odci.output_name: _MaterializableFeatureFunction(
                ff_for_serving=ff_for_serving_by_udf_name[odci.udf_name],
                odci=odci,
                fn=fns_by_udf_name[odci.udf_name],
            )
            for odci in on_demand_column_infos
        }

    def _convert_output_types(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Casts pandas DataFrame `df` to the expected output data types of FeatureFunctions.
        Expects all computed FeatureFunction outputs to be columns in `df`.

        :param df: pandas DataFrame with uncasted FeatureFunctions outputs.
        :return: pandas DataFrame casted to data types specified by FeatureFunction outputs.
        """
        spark_schema = [
            (output_name, self.materializable_features[output_name].return_data_type)
            for output_name in df.columns
        ]
        return feature_function_type_utils.cast_pandas_df(
            df=df, spark_schema=spark_schema
        )



""" Defines the type conversion classes for brickstore HTTP input JSONs and responses.
"""

from databricks.feature_store.utils.pandas_type_utils import (
    PandasArrayTypeConverter,
    PandasBase64BinaryTypeConverter,
    PandasDoubleTypeConverter,
    PandasFloatTypeConverter,
    PandasIntTypeConverter,
    PandasIsoDateTypeConverter,
    PandasIsoTimestampHttpTypeUsConverter,
    PandasMapTypeConverter,
    PandasObjectBooleanConverter,
    PandasShortTypeConverter,
    PandasStringDecimalTypeConverter,
    PandasStringLongTypeConverter,
    PandasStringTypeConverter,
)
from databricks.ml_features_common.entities.data_type import DataType
from databricks.ml_features_common.utils.converter_utils import ConverterFactory

"""
All supported converters.
"""
BASIC_DATA_TYPE_CONVERTERS = {
    DataType.SHORT: PandasShortTypeConverter,
    DataType.INTEGER: PandasIntTypeConverter,
    DataType.LONG: PandasStringLongTypeConverter,
    DataType.FLOAT: PandasFloatTypeConverter,
    DataType.DOUBLE: PandasDoubleTypeConverter,
    DataType.BOOLEAN: PandasObjectBooleanConverter,
    DataType.STRING: PandasStringTypeConverter,
    DataType.TIMESTAMP: PandasIsoTimestampHttpTypeUsConverter,
    DataType.DATE: PandasIsoDateTypeConverter,
    DataType.BINARY: PandasBase64BinaryTypeConverter,
    DataType.DECIMAL: PandasStringDecimalTypeConverter,
}

COMPLEX_DATA_TYPE_CONVERTERS = {
    DataType.ARRAY: PandasArrayTypeConverter,
    DataType.MAP: PandasMapTypeConverter,
}

BRICKSTORE_HTTP_DATA_TYPE_CONVERTER_FACTORY = ConverterFactory(
    BASIC_DATA_TYPE_CONVERTERS, COMPLEX_DATA_TYPE_CONVERTERS
)



import base64
import os
from typing import List

from databricks.ml_features_common.entities._feature_store_object import (
    _FeatureStoreObject,
)
from databricks.ml_features_common.entities.data_type import DataType
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    FeatureFunctionForServing as ProtoFeatureFunctionForServing,
)
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    FeatureFunctionParameterInfo as ProtoFeatureFunctionParameterInfo,
)
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    FeatureFunctionsForServing as ProtoFeatureFunctionsForServing,
)


class FeatureFunctionParameterInfo(_FeatureStoreObject):
    def __init__(self, type_text: str, type_json: str, name: str, type_name: DataType):
        self._type_text = type_text
        self._type_json = type_json
        self._name = name
        self._type_name = type_name

    @property
    def type_text(self) -> str:
        return self._type_text

    @property
    def type_json(self) -> str:
        return self._type_json

    @property
    def name(self) -> str:
        return self._name

    @property
    def type_name(self) -> DataType:
        return self._type_name

    @classmethod
    def from_proto(cls, the_proto: ProtoFeatureFunctionParameterInfo):
        return cls(
            type_text=the_proto.type_text,
            type_json=the_proto.type_json,
            name=the_proto.name,
            type_name=the_proto.type_name,
        )


class FeatureFunctionForServing(_FeatureStoreObject):
    def __init__(
        self,
        full_name: str,
        input_params: List[FeatureFunctionParameterInfo],
        data_type: DataType,
        full_data_type: str,
        routine_definition: str,
    ):
        self._full_name = full_name
        self._input_params = input_params
        self._data_type = data_type
        self._full_data_type = full_data_type
        self._routine_definition = routine_definition

    @property
    def full_name(self) -> str:
        return self._full_name

    @property
    def input_params(self) -> List[FeatureFunctionParameterInfo]:
        return self._input_params

    @property
    def data_type(self) -> DataType:
        return self._data_type

    @property
    def full_data_type(self) -> str:
        return self._full_data_type

    @property
    def routine_definition(self) -> str:
        return self._routine_definition

    @classmethod
    def from_proto(cls, the_proto: ProtoFeatureFunctionForServing):
        return cls(
            full_name=the_proto.full_name,
            input_params=[
                FeatureFunctionParameterInfo.from_proto(input_param)
                for input_param in the_proto.input_params
            ],
            data_type=the_proto.data_type,
            full_data_type=the_proto.full_data_type,
            routine_definition=the_proto.routine_definition,
        )


class FeatureFunctionsForServing(_FeatureStoreObject):
    DATA_FILE = "feature_functions_for_serving.dat"

    def __init__(self, ff_for_serving: List[FeatureFunctionForServing]):
        self._ff_for_serving = ff_for_serving

    @property
    def ff_for_serving(
        self,
    ) -> List[FeatureFunctionForServing]:
        return self._ff_for_serving

    @classmethod
    def from_proto(cls, the_proto: ProtoFeatureFunctionsForServing):
        ff_for_serving_objs = [
            FeatureFunctionForServing.from_proto(ff_for_serving)
            for ff_for_serving in the_proto.ff_for_serving
        ]
        return cls(ff_for_serving=ff_for_serving_objs)

    @classmethod
    def load(cls, path: str):
        """
        Loads a FeatureFunctionsForServing object from a base64 encoded proto representation.
        """
        proto = ProtoFeatureFunctionsForServing()
        with open(os.path.join(path, cls.DATA_FILE), "rb") as f:
            proto.ParseFromString(base64.b64decode(f.read()))
        return cls.from_proto(proto)



import json
from abc import ABC, abstractmethod
from typing import Dict, Union

import numpy as np

from databricks.ml_features_common.entities.online_feature_table import (
    FeatureDetails,
    PrimaryKeyDetails,
)
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    DataType as ProtoDataType,
)


# Abstract base class for converters
class Converter(ABC):
    pass


# Helper functions for converter decorators
def __validate_value(converter, args, kwargs):
    """
    Validate the value parameter passed into the converter function.
    :param converter: Static or non-static function.
    :param args: Positional arguments passed into the converter function.
    :param kwargs: Keyword arguments passed into the converter function.
    :return: True if value is valid; otherwise, False. If value is valid, returns the value as tuple
    """

    # If it is a stateful converter then first argument will always be self
    if len(args) > 0 and isinstance(args[0], Converter):
        non_self_args = args[1:]
    else:
        non_self_args = args

    # Input is invalid if both positional arguments (args) and keyword arguments (kwargs) are provided,
    # or if keyword arguments (kwargs) are provided with keyword other than value,
    # as it becomes ambiguous to determine the intended value.
    if len(non_self_args) + len(kwargs) != 1 or (
        len(kwargs) == 1 and "value" not in kwargs
    ):
        return False, None

    if "value" in kwargs:
        return True, kwargs["value"]
    else:
        return True, non_self_args[0]


def return_if_none(converter):
    """
    Decorator to return None early if the value parameter passed into converter function is None.
    :param converter: Static or non-static function.
    :return: None if input is None; otherwise, converted value.
    """

    def inner(*args, **kwargs):
        # This function expects either one positional argument or one keyword argument with name
        # "value".
        # Perform validation on kwargs and args before applying the None logic, this way if the user
        # incorrectly passes in wrong arguments we can rely on the underlying logic of converter
        # function to catch the errors.
        is_valid_value, value = __validate_value(converter, args, kwargs)
        if is_valid_value and value is None:
            return None

        return converter(*args, **kwargs)

    return inner


def return_if_nan(converter):
    """
    Decorator to return np.nan if the value parameter passed into the converter function is np.nan.
    This is used for the numpy.int16/32/64 types, as np.nan is not a valid constructor input for int types.

    TODO (ML-20967): Determine what represents a empty value in online lookup and update this decorator.
    """

    def inner(*args, **kwargs):
        """
        We currently use np.nan to represent missing values from the online store.
        np.nan is a float type and can't be coerced to a np.int16/32/64.
        In ML nan is usually used to represent missing values.
        """
        # This function expects either one positional argument or one keyword argument with name
        # "value".
        # Perform validation on kwargs and args before applying the Nan logic, this way if the user
        # incorrectly passes in wrong arguments we can rely on the underlying logic of converter
        # function to catch the errors.
        is_valid_value, value = __validate_value(converter, args, kwargs)
        if is_valid_value and isinstance(value, float) and np.isnan(value):
            return np.nan

        return converter(*args, **kwargs)

    return inner


class ConverterFactory:
    """
    ConverterFactory exposes a single factory method `get_converter` to generate a converter
    for the data type of the given feature/primary key.
    There should be a ConverterFactory instance defined for each online store.
    See `dynamodb_type_utils.py` for example usage.
    """

    def __init__(self, basic_datatype_converters, complex_datatype_converters):
        # Both inputs are dictionary of data type -> converter mapping for this online store.
        # See `dynamodb_type_utils.py` for example usage.
        self._basic_datatype_converters = basic_datatype_converters
        self._complex_datatype_converters = complex_datatype_converters

    def get_converter(self, details: Union[FeatureDetails, PrimaryKeyDetails]):
        # Looks up the data type converter subclass based on the input FeatureDetails or PrimaryKeyDetails.
        data_type_details = (
            json.loads(details.data_type_details)
            if isinstance(details, FeatureDetails) and details.data_type_details
            else None
        )
        return self._get_converter_detailed(
            data_type=details.data_type, details=data_type_details
        )

    def _get_converter_detailed(self, data_type, details: Union[str, Dict]):
        if data_type in self._basic_datatype_converters:
            return self._basic_datatype_converters[data_type]
        elif data_type in self._complex_datatype_converters and details:
            return self._complex_datatype_converters[data_type](
                details, self._get_converter_detailed
            )
        raise ValueError(
            f"Internal Error: Unsupported data type: {ProtoDataType.Name(data_type)}, details='{details}'"
        )



"""
Global registry for metrics recorder implementations.

This module manages the Feature Store metrics recorder lifecycle:
1. During initialization, model server/serving scheduler registers their metrics
   implementation (e.g., PrometheusMetricsRecorder) via register_metrics_recorder()
2. The lookup client retrieves the recorder via get_metrics_recorder() to emit metrics
3. The recorder wraps the implementation with a Feature Store-specific prefix

Thread safety: The global recorder is shared across all greenlets in the same process.
This is initialized once during worker startup (post_fork hook in model server) and
remains constant for the worker's lifetime.

No-op fallback: If no recorder is registered, a _NoOpRecorder is used to ensure
zero overhead in environments without metrics support (notebooks, local dev).
"""

import os
from typing import Dict, Optional

from databricks.feature_store.feature_lookup_version import VERSION
from databricks.feature_store.metrics.metrics_recorder_interface import (
    MetricsRecorder,
    _NoOpRecorder,
)
from databricks.feature_store.utils.lookup_client_envvars import (
    ENABLE_FEATURE_LOOKUP_METRICS_RECORDER_ENV,
)

# Singleton no-op recorder instance used as the default when no metrics are configured.
# Ensures zero overhead in environments without metrics support (notebooks, local dev).
_NOOP_RECORDER = _NoOpRecorder()

# Global metrics recorder shared across all greenlets in the same process.
# Initialized once during worker startup (via register_metrics_recorder) and remains
# constant for the worker's lifetime. Defaults to _NOOP_RECORDER until explicitly set.
# Thread safety: This is set once at initialization and then only read, avoiding race conditions.
_GLOBAL_RECORDER: MetricsRecorder = _NOOP_RECORDER


# Metric prefix. All metric names will be prefixed with this.
METRICS_PREFIX = "feature_lookup_client_"

# ==== Metrics definitions ====
# Latency added by lookup client. It doesn't include the latency of the raw model's prediction.
# This is measured from the time when lookup client logic receives the request to the time
# BEFORE calling the raw model predict method.
METRIC_FEATURE_STORE_LATENCY = METRICS_PREFIX + "feature_store_latency_ms"
# Latency measured for the raw model predict.
# feature_store_latency_ms + raw_model_predict_latency_ms = prediction_latency measured
# by the scoring server.
METRIC_RAW_MODEL_PREDICT_LATENCY = METRICS_PREFIX + "raw_model_predict_latency_ms"
# Postgres query latency. This is a histogram of latency for each individual query to the Postgres
# database. It does not include the acquisition of a connection to the database. Each request may
# involve multiple Postgres queries.
METRIC_POSTGRES_QUERY_LATENCY = METRICS_PREFIX + "postgres_query_latency_ms"
# Connection latency. This is the latency of acquiring a connection to the Postgres database.
METRIC_POSTGRES_CONNECTION_ACQUISITION_LATENCY = (
    METRICS_PREFIX + "postgres_connection_acquisition_latency_ms"
)
# Errors in Feature Store.
METRIC_FEATURE_STORE_ERROR_COUNT = METRICS_PREFIX + "feature_store_error_count"

# === Label definitions ====
# Lookup client version
METRIC_LABEL_LOOKUP_CLIENT_VERSION = "lookup_client_version"
# Online store type. Values are defined in databricks.ml_features_common.entities.store_type
METRIC_LABEL_ONLINE_STORE_TYPE = "online_store_type"
# Label indicating the source of the error. Values are "raw_model" or "feature_lookup".
METRIC_LABEL_ERROR_SOURCE = "error_source"


class FeatureStoreMetricsRecorder:
    """
    Wrapper around a MetricsRecorder that adds Feature Store-specific prefixes.

    This class wraps the runtime-provided metrics recorder (e.g., PrometheusMetricsRecorder)
    and guards the recording of metrics by the ENABLE_FEATURE_LOOKUP_METRICS_RECORDER_ENV
    environment variable.
    """

    def __init__(self, metrics_recorder: MetricsRecorder):
        """
        Initialize the Feature Store metrics recorder wrapper.

        :param metrics_recorder: The underlying MetricsRecorder implementation to wrap
        """
        self.base_labels = {
            METRIC_LABEL_LOOKUP_CLIENT_VERSION: VERSION,
        }
        self._metrics_recorder = metrics_recorder
        self._metrics_recorder_flag_enabled = (
            os.getenv(ENABLE_FEATURE_LOOKUP_METRICS_RECORDER_ENV) == "true"
        )

    def record_counter(
        self, name: str, value: float, labels: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Record a counter metric (monotonically increasing value).

        :param name: Metric name without prefix (e.g., "function_execution_count")
        :param value: Value to add to the counter (typically 1)
        :param labels: Optional label dimensions (e.g., {"table": "users", "status": "success"})
        """
        if not self._metrics_recorder_flag_enabled:
            return
        all_labels = {**self.base_labels, **(labels or {})}
        self._metrics_recorder.record_counter(name, value, all_labels)

    def record_histogram(
        self, name: str, value: float, labels: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Record a histogram metric (for distributions of values).

        :param name: Metric name without prefix (e.g., "postgres_query_latency_ms")
        :param value: Observed value (e.g., latency in milliseconds, payload size in bytes)
        :param labels: Optional label dimensions (e.g., {"table": "users", "source": "online_store"})
        """
        if not self._metrics_recorder_flag_enabled:
            return
        all_labels = {**self.base_labels, **(labels or {})}
        self._metrics_recorder.record_histogram(name, value, all_labels)


def register_metrics_recorder(recorder: MetricsRecorder):
    """
    Register a metrics recorder implementation.

    Model scoring server detects the existence of this method and uses it to register the recorder.
    This is called during worker initialization (post_fork hook) after the Feature Store wrapper
    model is initialized. The recorder is stored in a global variable.

    :param recorder: Implementation of MetricsRecorder (e.g., PrometheusMetricsRecorder)
    """
    global _GLOBAL_RECORDER
    wrapped_recorder = FeatureStoreMetricsRecorder(recorder)
    _GLOBAL_RECORDER = wrapped_recorder


def get_metrics_recorder() -> MetricsRecorder:
    """
    Get the current metrics recorder.

    Returns no-op recorder if none registered.

    :return: MetricsRecorder implementation or _NoOpRecorder
    """
    return _GLOBAL_RECORDER



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: feature.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    4,
    '',
    'feature.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rfeature.proto\x12\x12\x66\x65\x61turestorecommon\x1a\x15scalapb/scalapb.proto\"\x1f\n\x07\x46\x65\x61ture\x12\x14\n\x0c\x66\x65\x61ture_name\x18\x01 \x01(\tBq\n.com.databricks.proto.featurestorecommon.jprotoB\x0c\x46\x65\x61tureProtoP\x01\xa0\x01\x01\xe2?+\n\'com.databricks.proto.featurestorecommon\x10\x01')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feature_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n.com.databricks.proto.featurestorecommon.jprotoB\014FeatureProtoP\001\240\001\001\342?+\n\'com.databricks.proto.featurestorecommon\020\001'
  _globals['_FEATURE']._serialized_start=60
  _globals['_FEATURE']._serialized_end=91
# @@protoc_insertion_point(module_scope)
